# coding: utf-8

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import pytest
from datetime import datetime, timedelta

from devtools_testutils import AzureRecordedTestCase, recorded_by_proxy

from azure.data.tables import (
    ResourceTypes,
    AccountSasPermissions,
    TableRetentionPolicy,
    UpdateMode,
    TableAccessPolicy,
    TableAnalyticsLogging,
    TableMetrics,
    TableServiceClient,
    TableItem,
    generate_account_sas,
    ResourceTypes,
)
from azure.core.credentials import AzureSasCredential
from azure.core.exceptions import ResourceExistsError

from _shared.testcase import TableTestCase
from preparers import tables_decorator, tables_decorator

# ------------------------------------------------------------------------------


class TestTable(AzureRecordedTestCase, TableTestCase):
    @tables_decorator
    @recorded_by_proxy
    def test_create_properties(self, tables_storage_account_name, tables_primary_storage_account_key):
        # # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table_name = self._get_table_reference()
        # Act
        created = ts.create_table(table_name)

        # Assert
        assert created.table_name == table_name

        properties = ts.get_service_properties()
        ts.set_service_properties(analytics_logging=TableAnalyticsLogging(write=True))
        # have to wait for return to service
        p = ts.get_service_properties()
        # have to wait for return to service
        ts.set_service_properties(
            minute_metrics=TableMetrics(
                enabled=True, include_apis=True, retention_policy=TableRetentionPolicy(enabled=True, days=5)
            )
        )

        ps = ts.get_service_properties()
        ts.delete_table(table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_create_table(self, tables_storage_account_name, tables_primary_storage_account_key):
        # # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table_name = self._get_table_reference()

        # Act
        table = ts.get_table_client(table_name)
        created = table.create_table()

        # Assert
        assert created.name == table_name
        ts.delete_table(table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_create_table_fail_on_exist(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table_name = self._get_table_reference()

        # Act
        created = ts.create_table(table_name)
        with pytest.raises(ResourceExistsError):
            ts.create_table(table_name)

        name_filter = "TableName eq '{}'".format(table_name)
        existing = list(ts.query_tables(name_filter))

        # Assert
        assert created is not None
        ts.delete_table(table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_query_tables_per_page(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table_name = "mytable"

        for i in range(5):
            ts.create_table(table_name + str(i))

        query_filter = "TableName eq 'mytable0' or TableName eq 'mytable1' or TableName eq 'mytable2'"
        table_count = 0
        page_count = 0
        for table_page in ts.query_tables(query_filter, results_per_page=2).by_page():

            temp_count = 0
            for table in table_page:
                temp_count += 1
            assert temp_count <= 2
            page_count += 1
            table_count += temp_count

        assert page_count == 2
        assert table_count == 3

        self._delete_all_tables(ts)

    @tables_decorator
    @recorded_by_proxy
    def test_create_table_if_exists(self, tables_storage_account_name, tables_primary_storage_account_key):
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table_name = self._get_table_reference()

        t0 = ts.create_table(table_name)
        t1 = ts.create_table_if_not_exists(table_name)

        assert t0 is not None
        assert t1 is not None
        assert t0.table_name == t1.table_name
        ts.delete_table(table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_create_table_if_exists_new_table(self, tables_storage_account_name, tables_primary_storage_account_key):
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table_name = self._get_table_reference()

        t = ts.create_table_if_not_exists(table_name)

        assert t is not None
        assert t.table_name == table_name
        ts.delete_table(table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_query_tables(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        t = self._create_table(ts)

        # Act
        tables = list(ts.list_tables())

        # Assert
        for table_item in tables:
            assert isinstance(table_item, TableItem)

        assert tables is not None
        assert len(tables) >= 1
        assert tables[0] is not None

        self._delete_all_tables(ts)

    @tables_decorator
    @recorded_by_proxy
    def test_query_tables_with_filter(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        t = self._create_table(ts)

        # Act
        name_filter = "TableName eq '{}'".format(t.table_name)
        tables = list(ts.query_tables(name_filter))

        for table_item in tables:
            assert isinstance(table_item, TableItem)

        # Assert
        assert tables is not None
        assert len(tables) == 1
        ts.delete_table(t.table_name)

        self._delete_all_tables(ts)

    @tables_decorator
    @recorded_by_proxy
    def test_query_tables_with_num_results(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        prefix = "listtable"
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table_list = []
        for i in range(0, 4):
            self._create_table(ts, prefix + str(i), table_list)

        # Act
        small_page = []
        big_page = []
        for s in next(ts.list_tables(results_per_page=3).by_page()):
            small_page.append(s)
            assert s.name.startswith(prefix)
        for t in next(ts.list_tables().by_page()):
            big_page.append(t)
            assert t.name.startswith(prefix)

        # Assert
        assert len(small_page) == 3
        assert len(big_page) >= 4

        self._delete_all_tables(ts)

    @tables_decorator
    @recorded_by_proxy
    def test_query_tables_with_marker(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        prefix = "listtable"
        table_names = []
        for i in range(0, 4):
            self._create_table(ts, prefix + str(i), table_names)

        # Act
        generator1 = ts.list_tables(results_per_page=2).by_page()
        next(generator1)
        generator2 = ts.list_tables(results_per_page=2).by_page(continuation_token=generator1.continuation_token)
        next(generator2)

        tables1 = generator1._current_page
        tables2 = generator2._current_page

        # Assert
        assert len(tables1) == 2
        assert len(tables2) == 2
        assert tables1 != tables2

        self._delete_all_tables(ts)

    @tables_decorator
    @recorded_by_proxy
    def test_delete_table_with_existing_table(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table = self._create_table(ts)

        # Act
        deleted = ts.delete_table(table_name=table.table_name)
        existing = list(ts.query_tables("TableName eq '{}'".format(table.table_name)))

        # Assert
        assert deleted is None
        assert len(existing) == 0

    @tables_decorator
    @recorded_by_proxy
    def test_delete_table_with_non_existing_table_fail_not_exist(
        self, tables_storage_account_name, tables_primary_storage_account_key
    ):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table_name = self._get_table_reference()
        ts.delete_table(table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_get_table_acl(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        url = self.account_url(tables_storage_account_name, "table")
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table = self._create_table(ts)
        try:
            # Act
            acl = table.get_table_access_policy()

            # Assert
            assert acl is not None
            assert len(acl) == 0
        finally:
            ts.delete_table(table.table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_set_table_acl_with_empty_signed_identifiers(
        self, tables_storage_account_name, tables_primary_storage_account_key
    ):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")

        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table = self._create_table(ts)
        try:
            # Act
            table.set_table_access_policy(signed_identifiers={})

            # Assert
            acl = table.get_table_access_policy()
            assert acl is not None
            assert len(acl) == 0
        finally:
            ts.delete_table(table.table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_set_table_acl_with_empty_signed_identifier(
        self, tables_storage_account_name, tables_primary_storage_account_key
    ):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")

        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table = self._create_table(ts)
        try:
            dt = datetime(2021, 6, 8, 2, 10, 9)
            signed_identifiers = {
                "null": None,
                "empty": TableAccessPolicy(start=None, expiry=None, permission=None),
                "partial": TableAccessPolicy(permission="r"),
                "full": TableAccessPolicy(start=dt, expiry=dt, permission="r"),
            }
            table.set_table_access_policy(signed_identifiers)
            acl = table.get_table_access_policy()
            assert acl is not None
            assert len(acl) == 4
            assert acl["null"] is None
            assert acl["empty"] is None
            assert acl["partial"] is not None
            assert acl["partial"].permission == "r"
            assert acl["partial"].expiry is None
            assert acl["partial"].start is None
            assert acl["full"] is not None
            assert acl["full"].permission == "r"
            self._assert_policy_datetime(dt, acl["full"].expiry)
            self._assert_policy_datetime(dt, acl["full"].start)

            signed_identifiers.pop("empty")
            signed_identifiers["partial"] = None

            table.set_table_access_policy(signed_identifiers)
            acl = table.get_table_access_policy()
            assert acl is not None
            assert len(acl) == 3
            assert "empty" not in acl
            assert acl["null"] is None
            assert acl["partial"] is None
            assert acl["full"] is not None
            assert acl["full"].permission == "r"
            self._assert_policy_datetime(dt, acl["full"].expiry)
            self._assert_policy_datetime(dt, acl["full"].start)
        finally:
            ts.delete_table(table.table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_set_table_acl_with_signed_identifiers(
        self, tables_storage_account_name, tables_primary_storage_account_key
    ):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")

        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table = self._create_table(ts)
        client = ts.get_table_client(table_name=table.table_name)

        # Act
        start = datetime(2021, 6, 8, 2, 10, 9) - timedelta(minutes=5)
        expiry = datetime(2021, 6, 8, 2, 10, 9) + timedelta(hours=1)
        identifiers = {"testid": TableAccessPolicy(start=start, expiry=expiry, permission="r")}
        try:
            client.set_table_access_policy(signed_identifiers=identifiers)
            # Assert
            acl = client.get_table_access_policy()
            assert acl is not None
            assert len(acl) == 1
            assert acl.get("testid")
            self._assert_policy_datetime(start, acl["testid"].start)
            self._assert_policy_datetime(expiry, acl["testid"].expiry)
            assert acl["testid"].permission == "r"
        finally:
            ts.delete_table(table.table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_set_table_acl_too_many_ids(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")

        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table = self._create_table(ts)
        try:
            # Act
            identifiers = {}
            for i in range(0, 6):
                identifiers["id{}".format(i)] = None

            # Assert
            with pytest.raises(ValueError):
                table.set_table_access_policy(signed_identifiers=identifiers)
        finally:
            ts.delete_table(table.table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_account_sas(self, tables_storage_account_name, tables_primary_storage_account_key):
        account_url = self.account_url(tables_storage_account_name, "table")
        tsc = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table = self._create_table(tsc)
        try:
            entity = {
                "PartitionKey": "test",
                "RowKey": "test1",
                "text": "hello",
            }
            table.upsert_entity(mode=UpdateMode.MERGE, entity=entity)

            entity["RowKey"] = "test2"
            table.upsert_entity(mode=UpdateMode.MERGE, entity=entity)

            token = AzureSasCredential(
                self.generate_sas(
                    generate_account_sas,
                    tables_primary_storage_account_key,
                    resource_types=ResourceTypes(object=True),
                    permission=AccountSasPermissions(read=True),
                    expiry=datetime.utcnow() + timedelta(hours=1),
                    start=datetime.utcnow() - timedelta(minutes=1),
                )
            )

            account_url = self.account_url(tables_storage_account_name, "table")

            service = TableServiceClient(credential=token, endpoint=account_url)

            # Act

            sas_table = service.get_table_client(table.table_name)
            entities = list(sas_table.list_entities())

            # Assert
            assert len(entities) == 2
            assert entities[0]["text"] == "hello"
            assert entities[1]["text"] == "hello"
        finally:
            tsc.delete_table(table.table_name)

    @tables_decorator
    @recorded_by_proxy
    def test_unicode_create_table_unicode_name(self, tables_storage_account_name, tables_primary_storage_account_key):
        account_url = self.account_url(tables_storage_account_name, "table")
        tsc = TableServiceClient(account_url, credential=tables_primary_storage_account_key)
        invalid_table_name = "啊齄丂狛狜"

        with pytest.raises(ValueError) as excinfo:
            tsc.create_table(invalid_table_name)
            assert (
                "Storage table names must be alphanumeric, cannot begin with a number, and must be between 3-63 characters long."
                in str(excinfo)
            )

    @tables_decorator
    @recorded_by_proxy
    def test_create_table_invalid_name(self, tables_storage_account_name, tables_primary_storage_account_key):
        account_url = self.account_url(tables_storage_account_name, "table")
        tsc = TableServiceClient(account_url, credential=tables_primary_storage_account_key)
        invalid_table_name = "my_table"

        with pytest.raises(ValueError) as excinfo:
            tsc.create_table(invalid_table_name)
            assert (
                "Storage table names must be alphanumeric, cannot begin with a number, and must be between 3-63 characters long."
                in str(excinfo)
            )



# coding: utf-8

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: sample_get_entity_etag_and_timestamp_async.py

DESCRIPTION:
    These samples demonstrate how to get etag and timestamp of an entity.

USAGE:
    python sample_get_entity_etag_and_timestamp_async.py

    Set the environment variables with your own values before running the sample:
    1) TABLES_STORAGE_ENDPOINT_SUFFIX - the Table service account URL suffix
    2) TABLES_STORAGE_ACCOUNT_NAME - the name of the storage account
    3) TABLES_PRIMARY_STORAGE_ACCOUNT_KEY - the storage account access key
"""
import os
import asyncio
from datetime import datetime
from uuid import uuid4, UUID
from dotenv import find_dotenv, load_dotenv
from azure.data.tables.aio import TableClient
from typing_extensions import TypedDict


class EntityType(TypedDict, total=False):
    PartitionKey: str
    RowKey: str
    text: str
    color: str
    price: float
    last_updated: datetime
    product_id: UUID
    inventory_count: int
    barcode: bytes


class SampleGetEntityMetadata(object):
    def __init__(self):
        load_dotenv(find_dotenv())
        self.access_key = os.getenv("TABLES_PRIMARY_STORAGE_ACCOUNT_KEY")
        self.endpoint_suffix = os.getenv("TABLES_STORAGE_ENDPOINT_SUFFIX")
        self.account_name = os.getenv("TABLES_STORAGE_ACCOUNT_NAME")
        self.endpoint = "{}.table.{}".format(self.account_name, self.endpoint_suffix)
        self.connection_string = "DefaultEndpointsProtocol=https;AccountName={};AccountKey={};EndpointSuffix={}".format(
            self.account_name, self.access_key, self.endpoint_suffix
        )
        self.table_name = "SampleGetEntityMetadataAsync"
        self.entity: EntityType = {
            "PartitionKey": "color",
            "RowKey": "brand",
            "text": "Marker",
            "color": "Purple",
            "price": 4.99,
            "last_updated": datetime.today(),
            "product_id": uuid4(),
            "inventory_count": 42,
            "barcode": b"135aefg8oj0ld58",  # cspell:disable-line
        }

    async def get_entity_metadata(self):
        async with TableClient.from_connection_string(self.connection_string, self.table_name) as table_client:
            # Prepare a table and add an entity into it
            await table_client.create_table()
            await table_client.create_entity(entity=self.entity)

            entity = await table_client.get_entity(self.entity["PartitionKey"], self.entity["RowKey"])
            print("Entity metadata:")
            print(f"etag: {entity.metadata['etag']}")
            print(f"timestamp: {entity.metadata['timestamp']}")

            # Cleanup
            await table_client.delete_table()


async def main():
    sample = SampleGetEntityMetadata()
    await sample.get_entity_metadata()


if __name__ == "__main__":
    asyncio.run(main())



# coding: utf-8

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: sample_copy_table.py

DESCRIPTION:
    These samples demonstrate how to copy tables between table and blob storage.

USAGE:
    python sample_copy_table.py

    Set the environment variables with your own values before running the sample:
    1) TABLES_STORAGE_ENDPOINT_SUFFIX - the Table storage service account URL suffix
    2) TABLES_STORAGE_ACCOUNT_NAME - the Tables storage account name
    3) TABLES_PRIMARY_STORAGE_ACCOUNT_KEY - the Tables storage account access key
    4) STORAGE_ACCOUNT_NAME - the blob storage account name
    5) STORAGE_ACCOUNT_KEY - the blob storage account key
"""
import json
import os
from azure.storage.blob import BlobServiceClient
from azure.data.tables import TableServiceClient
from datetime import datetime
from dotenv import find_dotenv, load_dotenv
from uuid import uuid4, UUID
from typing_extensions import TypedDict


class EntityType(TypedDict, total=False):
    PartitionKey: str
    RowKey: str
    text: str
    color: str
    price: float
    last_updated: datetime
    product_id: UUID
    inventory_count: int
    barcode: bytes


class EntityTypeJSONSerializable(TypedDict, total=False):
    PartitionKey: str
    RowKey: str
    text: str
    color: str
    price: float
    last_updated: str
    product_id: str
    inventory_count: int
    barcode: str


class CopyTableSamples(object):
    def __init__(self):
        load_dotenv(find_dotenv())
        self.access_key = os.getenv("TABLES_PRIMARY_STORAGE_ACCOUNT_KEY")
        self.endpoint_suffix = os.getenv("TABLES_STORAGE_ENDPOINT_SUFFIX")
        self.account_name = os.getenv("TABLES_STORAGE_ACCOUNT_NAME")
        self.table_connection_string = f"DefaultEndpointsProtocol=https;AccountName={self.account_name};AccountKey={self.access_key};EndpointSuffix={self.endpoint_suffix}"
        self.copy_to_blob_table_name = "copytoblobtablename" + uuid4().hex
        self.copy_to_table_table_name = "copytotabletablename" + uuid4().hex
        self.blob_account_name = os.getenv("STORAGE_ACCOUNT_NAME")
        self.blob_account_key = os.getenv("STORAGE_ACCOUNT_KEY")
        self.blob_connection_string = f"DefDefaultEndpointsProtocol=https;AccountName={self.blob_account_name};AccountKey={self.blob_account_key};EndpointSuffix=core.windows.net"
        self.blob_service_client = BlobServiceClient.from_connection_string(self.blob_connection_string)
        self.entity: EntityType = {
            "PartitionKey": "color",
            "text": "Marker",
            "color": "Purple",
            "price": 4.99,
            "last_updated": datetime.today(),
            "product_id": uuid4(),
            "inventory_count": 42,
            "barcode": b"135aefg8oj0ld58",  # cspell:disable-line
        }

    def copy_table_from_table_to_blob(self):
        print("Start to copy table from Tables table to Storage blob.")
        print(f"Table name: {self.copy_to_blob_table_name}")
        self._setup_table()
        try:
            self.container_client = self.blob_service_client.create_container(self.copy_to_blob_table_name)
            # Upload in-memory table data to a blob that stays in a container
            for entity in self.table_client.list_entities():
                # Convert type datetime, bytes, UUID values to string as they're not JSON serializable
                entity["last_updated"] = entity["last_updated"].isoformat()
                entity["product_id"] = entity["product_id"].hex
                entity["barcode"] = entity["barcode"].decode("utf-8")
                blob_name = f"{entity['PartitionKey']}{entity['RowKey']}"
                blob_client = self.blob_service_client.get_blob_client(self.copy_to_blob_table_name, blob_name)
                blob_client.upload_blob(json.dumps(entity))
            print("Done!")
        finally:
            self._tear_down()

    def copy_table_from_blob_to_table(self):
        print("Start to copy table from Storage blob to Tables table.")
        print("Table name: " + self.copy_to_blob_table_name)
        self._setup_blob()
        try:
            # Download entities from blob to memory
            # Note: when entities size is too big, may need to do copy by chunk
            blob_list = self.container_client.list_blobs()
            # Upload entities to a table
            table_service_client = TableServiceClient.from_connection_string(
                conn_str=self.table_connection_string, table_name=self.copy_to_table_table_name
            )
            self.table_client = table_service_client.get_table_client(self.copy_to_table_table_name)
            self.table_client.create_table()
            for blob in blob_list:
                blob_client = self.container_client.get_blob_client(blob)
                blob_stream_downloader = blob_client.download_blob()
                entity = json.loads(blob_stream_downloader.readall())
                # Convert values back to their original types
                entity["last_updated"] = datetime.fromisoformat(entity["last_updated"])
                entity["product_id"] = UUID(entity["product_id"])
                entity["barcode"] = entity["barcode"].encode("utf-8")
                self.table_client.upsert_entity(entity)
            print("Done!")
        finally:
            self._tear_down()

    def _setup_table(self):
        table_service_client = TableServiceClient.from_connection_string(
            conn_str=self.table_connection_string, table_name=self.copy_to_blob_table_name
        )
        self.table_client = table_service_client.get_table_client(self.copy_to_blob_table_name)
        self.table_client.create_table()
        for i in range(10):
            self.entity["RowKey"] = str(i)
            self.table_client.create_entity(self.entity)

    def _setup_blob(self):
        self.container_client = self.blob_service_client.create_container(self.copy_to_table_table_name)
        # Convert type datetime, bytes, UUID values to string as they're not JSON serializable
        entity_serializable: EntityTypeJSONSerializable = {
            "PartitionKey": self.entity["PartitionKey"],
            "text": self.entity["text"],
            "color": self.entity["color"],
            "price": self.entity["price"],
            "last_updated": self.entity["last_updated"].isoformat(),
            "product_id": self.entity["product_id"].hex,
            "inventory_count": 42,
            "barcode": self.entity["barcode"].decode("utf-8"),
        }
        for i in range(10):
            entity_serializable["RowKey"] = str(i)
            blob_name = f"{entity_serializable['PartitionKey']}{entity_serializable['RowKey']}"
            blob_client = self.blob_service_client.get_blob_client(self.copy_to_table_table_name, blob_name)
            blob_client.upload_blob(json.dumps(entity_serializable))

    def _tear_down(self):
        self.table_client.delete_table()
        self.container_client.delete_container()


if __name__ == "__main__":
    sample = CopyTableSamples()
    sample.copy_table_from_table_to_blob()
    sample.copy_table_from_blob_to_table()



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from ._models import AccessPolicy
from ._models import CorsRule
from ._models import GeoReplication
from ._models import Logging
from ._models import Metrics
from ._models import RetentionPolicy
from ._models import SignedIdentifier
from ._models import TableEntityQueryResponse
from ._models import TableProperties
from ._models import TableQueryResponse
from ._models import TableResponse
from ._models import TableResponseProperties
from ._models import TableServiceError
from ._models import TableServiceProperties
from ._models import TableServiceStats

from ._enums import GeoReplicationStatusType
from ._enums import OdataMetadataFormat
from ._enums import ResponseFormat
from ._patch import __all__ as _patch_all
from ._patch import *  # pylint: disable=unused-wildcard-import
from ._patch import patch_sdk as _patch_sdk

__all__ = [
    "AccessPolicy",
    "CorsRule",
    "GeoReplication",
    "Logging",
    "Metrics",
    "RetentionPolicy",
    "SignedIdentifier",
    "TableEntityQueryResponse",
    "TableProperties",
    "TableQueryResponse",
    "TableResponse",
    "TableResponseProperties",
    "TableServiceError",
    "TableServiceProperties",
    "TableServiceStats",
    "GeoReplicationStatusType",
    "OdataMetadataFormat",
    "ResponseFormat",
]
__all__.extend([p for p in _patch_all if p not in __all__])
_patch_sdk()



import functools
import inspect
import time

from azure.core.credentials import AzureNamedKeyCredential
from azure.core.exceptions import HttpResponseError
from devtools_testutils import PowerShellPreparer, is_live

CosmosPreparer = functools.partial(
    PowerShellPreparer,
    "tables",
    tables_cosmos_account_name="fakecosmosaccount",
    tables_primary_cosmos_account_key="fakecosmosaccountkey",
)

TablesPreparer = functools.partial(
    PowerShellPreparer,
    "tables",
    tables_storage_account_name="faketableaccount",
    tables_primary_storage_account_key="faketablesaccountkey",
)


def trim_kwargs_from_test_function(fn, kwargs):
    # the next function is the actual test function. the kwargs need to be trimmed so
    # that parameters which are not required will not be passed to it.
    if not getattr(fn, "__is_preparer", False):
        try:
            args, _, kw, _, _, _, _ = inspect.getfullargspec(fn)
        except AttributeError:
            args, _, kw, _ = inspect.getargspec(fn)  # pylint: disable=deprecated-method
        if kw is None:
            args = set(args)
            for key in [k for k in kwargs if k not in args]:
                del kwargs[key]


def tables_decorator(func, **kwargs):
    @TablesPreparer()
    def wrapper(*args, **kwargs):
        key = kwargs.pop("tables_primary_storage_account_key")
        name = kwargs.pop("tables_storage_account_name")
        key = AzureNamedKeyCredential(key=key, name=name)

        kwargs["tables_primary_storage_account_key"] = key
        kwargs["tables_storage_account_name"] = name

        trimmed_kwargs = {k: v for k, v in kwargs.items()}
        trim_kwargs_from_test_function(func, trimmed_kwargs)

        func(*args, **trimmed_kwargs)

    return wrapper


def cosmos_decorator(func, **kwargs):
    @CosmosPreparer()
    def wrapper(*args, **kwargs):
        key = kwargs.pop("tables_primary_cosmos_account_key")
        name = kwargs.pop("tables_cosmos_account_name")
        key = AzureNamedKeyCredential(key=key, name=name)

        kwargs["tables_primary_cosmos_account_key"] = key
        kwargs["tables_cosmos_account_name"] = name

        trimmed_kwargs = {k: v for k, v in kwargs.items()}
        trim_kwargs_from_test_function(func, trimmed_kwargs)

        EXPONENTIAL_BACKOFF = 1
        RETRY_COUNT = 0

        try:
            return func(*args, **trimmed_kwargs)
        except HttpResponseError as exc:
            if exc.status_code != 429:
                raise
            print("Retrying: {} {}".format(RETRY_COUNT, EXPONENTIAL_BACKOFF))
            while RETRY_COUNT < 6:
                if is_live():
                    time.sleep(EXPONENTIAL_BACKOFF)
                try:
                    return func(*args, **trimmed_kwargs)
                except HttpResponseError as exc:
                    print("Retrying: {} {}".format(RETRY_COUNT, EXPONENTIAL_BACKOFF))
                    EXPONENTIAL_BACKOFF **= 2
                    RETRY_COUNT += 1
                    if exc.status_code != 429 or RETRY_COUNT >= 6:
                        raise

    return wrapper



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import pytest

from devtools_testutils import AzureRecordedTestCase
from devtools_testutils.aio import recorded_by_proxy_async

from azure.data.tables.aio import TableServiceClient

from _shared.asynctestcase import AsyncTableTestCase
from async_preparers import cosmos_decorator_async

SERVICE_UNAVAILABLE_RESP_BODY = (
    '<?xml version="1.0" encoding="utf-8"?><StorageServiceStats><GeoReplication><Status'
    ">unavailable</Status><LastSyncTime></LastSyncTime></GeoReplication"
    "></StorageServiceStats> "
)

SERVICE_LIVE_RESP_BODY = (
    '<?xml version="1.0" encoding="utf-8"?><StorageServiceStats><GeoReplication><Status'
    ">live</Status><LastSyncTime>Wed, 19 Jan 2021 22:28:43 GMT</LastSyncTime></GeoReplication"
    "></StorageServiceStats> "
)


class TestTableServiceStatsCosmosAsync(AzureRecordedTestCase, AsyncTableTestCase):
    @staticmethod
    def override_response_body_with_unavailable_status(response):
        response.http_response.text = lambda _: SERVICE_UNAVAILABLE_RESP_BODY

    @staticmethod
    def override_response_body_with_live_status(response):
        response.http_response.text = lambda _: SERVICE_LIVE_RESP_BODY

    # --Test cases per service ---------------------------------------
    @pytest.mark.skip("JSON is invalid for cosmos")
    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_table_service_stats_f(self, tables_cosmos_account_name, tables_primary_cosmos_account_key):
        tsc = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        stats = await tsc.get_service_stats(raw_response_hook=self.override_response_body_with_live_status)
        self._assert_stats_default(stats)

    @pytest.mark.skip("JSON is invalid for cosmos")
    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_table_service_stats_when_unavailable(
        self, tables_cosmos_account_name, tables_primary_cosmos_account_key
    ):
        tsc = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        stats = await tsc.get_service_stats(raw_response_hook=self.override_response_body_with_unavailable_status)
        self._assert_stats_unavailable(stats)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from typing import Union, Any, Mapping, Optional, List, Tuple

from azure.core import MatchConditions
from azure.core.rest import HttpRequest

from ._common_conversion import _transform_patch_to_cosmos_post
from ._models import UpdateMode, TransactionOperation
from ._serialize import _add_entity_properties, _prepare_key, _get_match_condition
from ._entity import TableEntity
from ._generated.operations._operations import (
    build_table_insert_entity_request,
    build_table_merge_entity_request,
    build_table_update_entity_request,
    build_table_delete_entity_request,
)
from ._generated._configuration import AzureTableConfiguration
from ._generated.aio._configuration import AzureTableConfiguration as AsyncAzureTableConfiguration


EntityType = Union[TableEntity, Mapping[str, Any]]
OperationType = Union[TransactionOperation, str]
TransactionOperationType = Union[Tuple[OperationType, EntityType], Tuple[OperationType, EntityType, Mapping[str, Any]]]


class TableBatchOperations(object):
    """
    This is the class that is used for batch operations for the data tables
    service.

    The Tables service supports batch transactions on entities that are in the
    same table and belong to the same partition group. Multiple operations are
    supported within a single transaction. The batch can include at most 100
    entities, and its total payload may be no more than 4 MB in size.

    :ivar str table_name: The name of the table.
    :ivar requests: A list of :class:`~azure.core.rest.HttpRequest` in a batch.
    :vartype requests: list[~azure.core.rest.HttpRequest]
    """

    def __init__(
        self,
        config: Union[AzureTableConfiguration, AsyncAzureTableConfiguration],
        endpoint: str,
        table_name: str,
        is_cosmos_endpoint: bool = False,
    ) -> None:
        """Create TableClient from a Credential.

        :param config: An AzureTableConfiguration object.
        :type config: ~azure.data.tables._generated._configuration.AzureTableConfiguration
        :param endpoint: The primary account URL.
        :type endpoint: str
        :param table_name: The name of the Table to perform operations on.
        :type table_name: str
        :param is_cosmos_endpoint: True if the client endpoint is for Tables Cosmos. False if not. Default is False.
        :type is_cosmos_endpoint: bool
        """
        self._config = config
        self._base_url = endpoint
        self._is_cosmos_endpoint = is_cosmos_endpoint
        self.table_name = table_name

        self._partition_key: Optional[str] = None
        self.requests: List[HttpRequest] = []

    def __len__(self) -> int:
        return len(self.requests)

    def _verify_partition_key(self, entity: EntityType) -> None:
        if "PartitionKey" not in entity or "RowKey" not in entity:
            raise ValueError("PartitionKey and/or RowKey were not provided in entity")
        if self._partition_key is None:
            self._partition_key = entity["PartitionKey"]
        elif entity["PartitionKey"] != self._partition_key:
            raise ValueError("Partition Keys in the batch must all be the same.")

    def add_operation(self, operation: TransactionOperationType) -> None:
        """Add a single operation to a batch.

        :param operation: An operation include operation type and entity, may with kwargs.
        :type operation: A tuple of ~azure.data.tables.TransactionOperation or str, and
            ~azure.data.tables.TableEntity or Mapping[str, Any]. Or a tuple of
            ~azure.data.tables.TransactionOperation or str, and
            ~azure.data.tables.TableEntity or Mapping[str, Any], and Mapping[str, Any]
        :return: None
        """
        if len(operation) == 3:
            operation_type, entity, kwargs = operation  # type: ignore[misc]
        else:
            operation_type, entity, kwargs = operation[0], operation[1], {}
        try:
            getattr(self, operation_type.lower())(entity, **kwargs)
        except AttributeError as exc:
            raise ValueError(f"Unrecognized operation: {operation_type}") from exc

    def create(self, entity: EntityType, **kwargs) -> None:
        """Adds an insert operation to the current batch.

        :param entity: The properties for the table entity.
        :type entity: ~azure.data.tables.TableEntity or dict[str, Any]
        :return: None
        :raises ValueError:

        .. admonition:: Example:

            .. literalinclude:: ../samples/sample_batching.py
                :start-after: [START batching]
                :end-before: [END batching]
                :language: python
                :dedent: 8
                :caption: Creating and adding an entity to a Table
        """
        self._verify_partition_key(entity)
        entity = _add_entity_properties(entity)
        request = build_table_insert_entity_request(
            table=self.table_name, json=entity, version=self._config.version, **kwargs
        )
        request.url = self._base_url + request.url
        self.requests.append(request)

    def update(
        self,
        entity: EntityType,
        mode: Union[str, UpdateMode] = UpdateMode.MERGE,
        *,
        etag: Optional[str] = None,
        match_condition: Optional[MatchConditions] = None,
        **kwargs,
    ) -> None:
        """Adds an update operation to the current batch.

        :param entity: The properties for the table entity.
        :type entity: ~azure.data.tables.TableEntity or dict[str, Any]
        :param mode: Merge or Replace entity
        :type mode: ~azure.data.tables.UpdateMode
        :keyword etag: Etag of the entity.
        :paramtype etag: str or None
        :keyword match_condition: The match condition to use upon the etag.
        :paramtype match_condition: ~azure.core.MatchConditions or None
        :return: None
        :raises ValueError:

        .. admonition:: Example:

            .. literalinclude:: ../samples/sample_batching.py
                :start-after: [START batching]
                :end-before: [END batching]
                :language: python
                :dedent: 8
                :caption: Creating and adding an entity to a Table
        """
        self._verify_partition_key(entity)
        entity = _add_entity_properties(entity)
        partition_key = _prepare_key(entity["PartitionKey"])
        row_key = _prepare_key(entity["RowKey"])

        if match_condition and not etag and isinstance(entity, TableEntity):
            if hasattr(entity, "metadata"):
                etag = entity.metadata.get("etag")
        match_condition = _get_match_condition(
            etag=etag, match_condition=match_condition or MatchConditions.Unconditionally
        )
        if mode == UpdateMode.REPLACE:
            request = build_table_update_entity_request(
                table=self.table_name,
                partition_key=partition_key,
                row_key=row_key,
                etag=etag,
                match_condition=match_condition,
                json=entity,
                version=self._config.version,
                **kwargs,
            )
        elif mode == UpdateMode.MERGE:
            request = build_table_merge_entity_request(
                table=self.table_name,
                partition_key=partition_key,
                row_key=row_key,
                etag=etag,
                match_condition=match_condition,
                json=entity,
                version=self._config.version,
                **kwargs,
            )
            if self._is_cosmos_endpoint:
                _transform_patch_to_cosmos_post(request)
        else:
            raise ValueError(f"Mode type '{mode}' is not supported.")

        request.url = self._base_url + request.url
        self.requests.append(request)

    def delete(
        self,
        entity: EntityType,
        *,
        etag: Optional[str] = None,
        match_condition: Optional[MatchConditions] = None,
        **kwargs,
    ) -> None:
        """Adds a delete operation to the current branch.

        :param entity: The properties for the table entity.
        :type entity: ~azure.data.tables.TableEntity or dict[str, Any]
        :keyword etag: Etag of the entity.
        :paramtype etag: str or None
        :keyword match_condition: The match condition to use upon the etag.
        :paramtype match_condition: ~azure.core.MatchConditions or None
        :return: None
        :raises: ValueError

        .. admonition:: Example:

            .. literalinclude:: ../samples/sample_batching.py
                :start-after: [START batching]
                :end-before: [END batching]
                :language: python
                :dedent: 8
                :caption: Creating and adding an entity to a Table
        """
        self._verify_partition_key(entity)
        if match_condition and not etag and isinstance(entity, TableEntity):
            etag = entity.metadata.get("etag")
        request = build_table_delete_entity_request(
            table=self.table_name,
            partition_key=_prepare_key(entity["PartitionKey"]),
            row_key=_prepare_key(entity["RowKey"]),
            etag=etag,  # type: ignore[arg-type] # Set None to skip checking etag.
            match_condition=_get_match_condition(
                etag=etag, match_condition=match_condition or MatchConditions.Unconditionally
            ),
            version=self._config.version,
            **kwargs,
        )
        request.url = self._base_url + request.url
        self.requests.append(request)

    def upsert(self, entity: EntityType, mode: Union[str, UpdateMode] = UpdateMode.MERGE, **kwargs) -> None:
        """Adds an upsert (merge or replace) operation to the batch.

        :param entity: The properties for the table entity.
        :type entity: ~azure.data.tables.TableEntity or dict[str, Any]
        :param mode: Merge or Replace entity
        :type mode: ~azure.data.tables.UpdateMode
        :return: None
        :raises: ValueError

        .. admonition:: Example:

            .. literalinclude:: ../samples/sample_batching.py
                :start-after: [START batching]
                :end-before: [END batching]
                :language: python
                :dedent: 8
                :caption: Creating and adding an entity to a Table
        """
        self._verify_partition_key(entity)

        # TODO: This ordering is backwards. Tracked as part of issue #26318
        partition_key = _prepare_key(entity["PartitionKey"])
        row_key = _prepare_key(entity["RowKey"])
        entity = _add_entity_properties(entity)

        if mode == UpdateMode.REPLACE:
            request = build_table_update_entity_request(
                table=self.table_name,
                partition_key=partition_key,
                row_key=row_key,
                json=entity,
                version=self._config.version,
                **kwargs,
            )
        elif mode == UpdateMode.MERGE:
            request = build_table_merge_entity_request(
                table=self.table_name,
                partition_key=partition_key,
                row_key=row_key,
                json=entity,
                version=self._config.version,
                **kwargs,
            )
            if self._is_cosmos_endpoint:
                _transform_patch_to_cosmos_post(request)
        else:
            raise ValueError(f"Mode type '{mode}' is not supported.")

        request.url = self._base_url + request.url
        self.requests.append(request)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from binascii import hexlify
from typing import Callable, Dict, Optional, Union, Any, Mapping, Type, Tuple
from uuid import UUID
from datetime import datetime
from math import isnan
from enum import Enum

from azure.core import MatchConditions

from ._entity import EdmType, TableEntity
from ._common_conversion import _encode_base64, _to_utc_datetime
from ._error import _ERROR_VALUE_TOO_LARGE, _ERROR_TYPE_NOT_SUPPORTED


def _get_match_condition(etag, match_condition):
    if match_condition == MatchConditions.IfNotModified:
        if not etag:
            raise ValueError("IfNotModified must be specified with etag.")
        return match_condition
    if match_condition == MatchConditions.Unconditionally:
        if etag:
            raise ValueError("Etag is not supported for an Unconditional operation.")
        return MatchConditions.IfPresent
    raise ValueError(f"Unsupported match condition: {match_condition}")


def _prepare_key(keyvalue: str) -> str:
    """Duplicate the single quote char to escape.
    :param keyvalue: A key value in table entity.
    :type keyvalue: str
    :return: A key value in table entity.
    :rtype: str
    """
    try:
        return keyvalue.replace("'", "''")
    except AttributeError as exc:
        raise TypeError("PartitionKey or RowKey must be of type string.") from exc


def _parameter_filter_substitution(parameters: Optional[Dict[str, str]], query_filter: str) -> str:
    """Replace user defined parameters in filter.

    :param parameters: User defined parameters
    :type parameters: dict[str, str]
    :param str query_filter: Filter for querying
    :return: A query filter replaced by user defined parameters.
    :rtype: str
    """
    if parameters:
        filter_strings = query_filter.split(" ")
        for index, word in enumerate(filter_strings):
            if word[0] == "@":
                val = parameters[word[1:]]
                if val in [True, False]:
                    filter_strings[index] = str(val).lower()
                elif isinstance(val, (float)):
                    filter_strings[index] = str(val)
                elif isinstance(val, int):
                    if val.bit_length() <= 32:
                        filter_strings[index] = str(val)
                    else:
                        filter_strings[index] = f"{str(val)}L"
                elif isinstance(val, datetime):
                    filter_strings[index] = f"datetime'{_to_utc_datetime(val)}'"
                elif isinstance(val, UUID):
                    filter_strings[index] = f"guid'{str(val)}'"
                elif isinstance(val, bytes):
                    v = str(hexlify(val))
                    v = v[2:-1]  # Python 3 adds a 'b' and quotations
                    filter_strings[index] = f"X'{v}'"
                else:
                    val = val.replace("'", "''")
                    filter_strings[index] = f"'{val}'"
        return " ".join(filter_strings)
    return query_filter


def _to_entity_binary(value):
    return EdmType.BINARY, _encode_base64(value)


def _to_entity_bool(value):
    return None, value


def _to_entity_datetime(value):
    if isinstance(value, str):
        # Pass a serialized datetime straight through
        return EdmType.DATETIME, value
    try:
        # Check is this is a 'round-trip' datetime, and if so
        # pass through the original value.
        if value.tables_service_value:
            return EdmType.DATETIME, value.tables_service_value
    except AttributeError:
        pass
    return EdmType.DATETIME, _to_utc_datetime(value)


def _to_entity_float(value):
    if isinstance(value, str):
        # Pass a serialized value straight through
        return EdmType.DOUBLE, value
    if isnan(value):
        return EdmType.DOUBLE, "NaN"
    if value == float("inf"):
        return EdmType.DOUBLE, "Infinity"
    if value == float("-inf"):
        return EdmType.DOUBLE, "-Infinity"
    return EdmType.DOUBLE, value


def _to_entity_guid(value):
    return EdmType.GUID, str(value)


def _to_entity_int32(value):
    value = int(value)
    if value >= 2**31 or value < -(2**31):
        raise TypeError(_ERROR_VALUE_TOO_LARGE.format(str(value), EdmType.INT32))
    return None, value


def _to_entity_int64(value):
    int_value = int(value)
    if int_value >= 2**63 or int_value < -(2**63):
        raise TypeError(_ERROR_VALUE_TOO_LARGE.format(str(value), EdmType.INT64))
    return EdmType.INT64, str(value)


def _to_entity_str(value):
    return EdmType.STRING, str(value)


def _to_entity_none(value):  # pylint: disable=unused-argument
    return None, None


# Conversion from Python type to a function which returns a tuple of the
# type string and content string.
_PYTHON_TO_ENTITY_CONVERSIONS: Dict[Type, Callable[[Any], Tuple[Optional[EdmType], Any]]] = {
    int: _to_entity_int32,
    bool: _to_entity_bool,
    datetime: _to_entity_datetime,
    float: _to_entity_float,
    UUID: _to_entity_guid,
    Enum: _to_entity_str,
    str: _to_entity_str,
    bytes: _to_entity_binary,
}

# cspell:ignore Odatatype

# Conversion from Edm type to a function which returns a tuple of the
# type string and content string. These conversions are only used when the
# full EdmProperty tuple is specified. As a result, in this case we ALWAYS add
# the Odatatype tag, even for field types where it's not necessary. This is why
# boolean and int32 have special processing below, as we would not normally add the
# Odatatype tags for these to keep payload size minimal.
# This is also necessary for CLI compatibility.
_EDM_TO_ENTITY_CONVERSIONS: Dict[EdmType, Callable[[Any], Tuple[Optional[EdmType], Any]]] = {
    EdmType.BINARY: _to_entity_binary,
    EdmType.BOOLEAN: lambda v: (EdmType.BOOLEAN, v),
    EdmType.DATETIME: _to_entity_datetime,
    EdmType.DOUBLE: _to_entity_float,
    EdmType.GUID: _to_entity_guid,
    EdmType.INT32: lambda v: (EdmType.INT32, _to_entity_int32(v)[1]),  # Still using the int32 validation
    EdmType.INT64: _to_entity_int64,
    EdmType.STRING: _to_entity_str,
}


def _add_entity_properties(source: Union[TableEntity, Mapping[str, Any]]) -> Dict[str, Any]:
    """Converts an entity object to json to send.
    The entity format is:
    {
       "Address":"Mountain View",
       "Age":23,
       "AmountDue":200.23,
       "CustomerCode@odata.type":"Edm.Guid",
       "CustomerCode":"c9da6455-213d-42c9-9a79-3e9149a57833",
       "CustomerSince@odata.type":"Edm.DateTime",
       "CustomerSince":"2008-07-10T00:00:00",
       "IsActive":true,
       "NumberOfOrders@odata.type":"Edm.Int64",
       "NumberOfOrders":"255",
       "PartitionKey":"my_partition_key",
       "RowKey":"my_row_key"
    }
    :param source: A table entity.
    :type source: ~azure.data.tables.TableEntity or Mapping[str, Any]
    :return: An entity with property's metadata in JSON format.
    :rtype: dict
    """

    properties = {}

    to_send = dict(source)  # shallow copy

    # set properties type for types we know if value has no type info.
    # if value has type info, then set the type to value.type
    for name, value in to_send.items():
        if value is None:
            continue
        mtype: Optional[EdmType] = None
        if isinstance(value, Enum):
            convert = _PYTHON_TO_ENTITY_CONVERSIONS[str]
            mtype, value = convert(value)
        elif isinstance(value, datetime):
            mtype, value = _to_entity_datetime(value)
        elif isinstance(value, tuple):
            if value[0] is None:
                continue
            convert = _EDM_TO_ENTITY_CONVERSIONS[EdmType(value[1])]
            mtype, value = convert(value[0])
        else:
            try:
                convert = _PYTHON_TO_ENTITY_CONVERSIONS[type(value)]
            except KeyError:
                raise TypeError(_ERROR_TYPE_NOT_SUPPORTED.format(type(value))) from None
            mtype, value = convert(value)

        # form the property node
        properties[name] = value
        if mtype:
            properties[name + "@odata.type"] = mtype.value

    # generate the entity_body
    return properties


def serialize_iso(attr: Optional[Union[str, datetime]]) -> Optional[str]:
    """Serialize Datetime object into ISO-8601 formatted string.

    :param attr: Object to be serialized.
    :type attr: str or ~datetime.datetime or None
    :return: A ISO-8601 formatted string or None
    :rtype: str or None
    :raises ValueError: When unable to serialize the input object.
    :raises TypeError: When ISO-8601 object is an invalid datetime object.
    """
    if not attr:
        return None
    if isinstance(attr, str):
        # Pass a string through unaltered
        return attr
    try:
        utc = attr.utctimetuple()
        if utc.tm_year > 9999 or utc.tm_year < 1:
            raise OverflowError("Hit max or min date")

        date = f"{utc.tm_year:04}-{utc.tm_mon:02}-{utc.tm_mday:02}T{utc.tm_hour:02}:{utc.tm_min:02}:{utc.tm_sec:02}"
        return date + "Z"
    except (ValueError, OverflowError) as err:
        raise ValueError("Unable to serialize datetime object.") from err
    except AttributeError as err:
        raise TypeError("ISO-8601 object must be valid datetime object.") from err



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import pytest

from devtools_testutils import AzureRecordedTestCase
from devtools_testutils.aio import recorded_by_proxy_async

from azure.core.exceptions import ResourceExistsError, HttpResponseError
from azure.data.tables.aio import TableServiceClient, TableClient

from _shared.asynctestcase import AsyncTableTestCase
from async_preparers import cosmos_decorator_async


class TestTableCosmosAsync(AzureRecordedTestCase, AsyncTableTestCase):
    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_create_table(self, tables_cosmos_account_name, tables_primary_cosmos_account_key):
        # Arrange
        ts = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        table_name = self._get_table_reference()

        # Act
        table = ts.get_table_client(table_name)
        created = await table.create_table()

        # Assert
        assert created.name == table_name

        await ts.delete_table(table_name=table_name)

    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_create_table_fail_on_exist(self, tables_cosmos_account_name, tables_primary_cosmos_account_key):
        # Arrange
        ts = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        table_name = self._get_table_reference()

        # Act
        created = await ts.create_table(table_name=table_name)
        with pytest.raises(ResourceExistsError):
            await ts.create_table(table_name=table_name)

        # Assert
        assert created
        await ts.delete_table(table_name=table_name)

    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_query_tables_per_page(self, tables_cosmos_account_name, tables_primary_cosmos_account_key):
        # Arrange
        ts = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )

        table_name = "myasynctable"

        for i in range(5):
            await ts.create_table(table_name + str(i))

        query_filter = "TableName eq 'myasynctable0' or TableName eq 'myasynctable1' or TableName eq 'myasynctable2'"
        table_count = 0
        page_count = 0
        async for table_page in ts.query_tables(query_filter, results_per_page=2).by_page():

            temp_count = 0
            async for table in table_page:
                temp_count += 1
            assert temp_count <= 2
            page_count += 1
            table_count += temp_count

        assert page_count == 2
        assert table_count == 3

        await self._delete_all_tables(tables_cosmos_account_name, tables_primary_cosmos_account_key)

    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_list_tables(self, tables_cosmos_account_name, tables_primary_cosmos_account_key):
        # Arrange
        ts = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        table = await self._create_table(ts)

        # Act
        tables = []
        async for t in ts.list_tables():
            tables.append(t)

        # Assert
        assert tables is not None
        assert len(tables) >= 1
        assert tables[0] is not None

    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_query_tables_with_filter(self, tables_cosmos_account_name, tables_primary_cosmos_account_key):
        # Arrange
        ts = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        table = await self._create_table(ts)

        # Act
        name_filter = "TableName eq '{}'".format(table.table_name)
        tables = []
        async for t in ts.query_tables(name_filter):
            tables.append(t)

        # Assert
        assert tables is not None
        assert len(tables) == 1
        await ts.delete_table(table.table_name)

    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_list_tables_with_num_results(self, tables_cosmos_account_name, tables_primary_cosmos_account_key):
        # Arrange
        await self._delete_all_tables(tables_cosmos_account_name, tables_primary_cosmos_account_key)
        prefix = "listtable"
        ts = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        table_list = []
        for i in range(0, 4):
            await self._create_table(ts, prefix + str(i), table_list)

        # Act
        all_tables = 0
        async for t in ts.list_tables():
            all_tables += 1

        small_page = 0
        async for page in ts.list_tables(results_per_page=3).by_page():
            page_size = 0
            async for table in page:
                page_size += 1
            assert page_size <= 3
            small_page += 1

        assert small_page == 2
        assert all_tables == 4

    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_list_tables_with_marker(self, tables_cosmos_account_name, tables_primary_cosmos_account_key):
        # Arrange
        ts = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        prefix = "listtable"
        table_names = []
        for i in range(0, 4):
            await self._create_table(ts, prefix + str(i), table_names)

        # Act
        generator1 = ts.list_tables(results_per_page=2).by_page()
        await generator1.__anext__()

        generator2 = ts.list_tables(results_per_page=2).by_page(continuation_token=generator1.continuation_token)
        await generator2.__anext__()

        tables1 = generator1._current_page
        tables2 = generator2._current_page

        tables1_len = 0
        async for _ in tables1:
            tables1_len += 1
        tables2_len = 0
        async for _ in tables2:
            tables2_len += 1

        # Assert
        assert tables1_len == 2
        assert tables2_len == 2
        assert tables1 != tables2

    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_delete_table_with_existing_table(
        self, tables_cosmos_account_name, tables_primary_cosmos_account_key
    ):
        # Arrange
        ts = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        table = await self._create_table(ts)

        # Act
        deleted = await ts.delete_table(table_name=table.table_name)

        # Assert
        assert deleted is None

    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_delete_table_with_non_existing_table_fail_not_exist(
        self, tables_cosmos_account_name, tables_primary_cosmos_account_key
    ):
        # Arrange
        ts = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        table_name = self._get_table_reference()
        await ts.delete_table(table_name)

    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_create_table_underscore_name(self, tables_cosmos_account_name, tables_primary_cosmos_account_key):
        # Arrange
        ts = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        table_name = "my_table"

        client = await ts.create_table(table_name)
        assert client.table_name == table_name

        await ts.delete_table(table_name)

    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_create_table_unicode_name(self, tables_cosmos_account_name, tables_primary_cosmos_account_key):
        # Arrange
        ts = TableServiceClient(
            self.account_url(tables_cosmos_account_name, "cosmos"), credential=tables_primary_cosmos_account_key
        )
        table_name = "啊齄丂狛狜"

        client = await ts.create_table(table_name)
        assert client.table_name == table_name

        await ts.delete_table(table_name)

    @cosmos_decorator_async
    @recorded_by_proxy_async
    async def test_client_with_url_ends_with_table_name(
        self, tables_cosmos_account_name, tables_primary_cosmos_account_key
    ):
        url = self.account_url(tables_cosmos_account_name, "cosmos")
        table_name = self.get_resource_name("mytable")
        invalid_url = url + "/" + table_name
        # test table client has the same table name as in url
        tc = TableClient(invalid_url, table_name, credential=tables_primary_cosmos_account_key)
        with pytest.raises(HttpResponseError) as exc:
            await tc.create_table()
        assert ("Request url is invalid") in str(exc.value)
        assert ("Please check your account URL.") in str(exc.value)
        # test table client has a different table name as in url
        table_name2 = self.get_resource_name("mytable2")
        tc2 = TableClient(invalid_url, table_name2, credential=tables_primary_cosmos_account_key)
        with pytest.raises(HttpResponseError) as exc:
            await tc2.create_table()
        assert ("Request url is invalid") in str(exc.value)
        assert ("Please check your account URL.") in str(exc.value)

        valid_tc = TableClient(url, table_name, credential=tables_primary_cosmos_account_key)
        await valid_tc.create_table()
        # test creating a table when it already exists
        with pytest.raises(HttpResponseError) as exc:
            await tc.create_table()
        assert ("Request url is invalid") in str(exc.value)
        assert ("Please check your account URL.") in str(exc.value)
        # test deleting a table when it already exists
        with pytest.raises(HttpResponseError) as exc:
            await tc.delete_table()
        assert ("Request url is invalid") in str(exc.value)
        assert ("Please check your account URL.") in str(exc.value)
        await valid_tc.delete_table()
        await valid_tc.close()
        await tc.close()
        await tc2.close()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from urllib.parse import urlparse
from typing import Optional, Union, overload, cast

from azure.core.credentials import TokenCredential, AzureSasCredential, AzureNamedKeyCredential
from azure.core.exceptions import ClientAuthenticationError
from azure.core.pipeline import PipelineResponse, PipelineRequest
from azure.core.pipeline.policies import BearerTokenCredentialPolicy, SansIOHTTPPolicy, AzureSasCredentialPolicy
from azure.core.pipeline.transport import AsyncHttpTransport

try:
    from yarl import URL  # cspell:disable-line
except ImportError:
    pass

from ._common_conversion import _sign_string
from ._error import _wrap_exception
from ._constants import STORAGE_OAUTH_SCOPE, COSMOS_OAUTH_SCOPE


class AzureSigningError(ClientAuthenticationError):
    """
    Represents a fatal error when attempting to sign a request.
    In general, the cause of this exception is user error. For example, the given account key is not valid.
    Please visit https://learn.microsoft.com/en-us/azure/storage/common/storage-create-storage-account for more info.
    """


class _HttpChallenge(object):  # pylint:disable=too-few-public-methods
    """Represents a parsed HTTP WWW-Authentication Bearer challenge from a server.

    :param challenge: The WWW-Authenticate header of the challenge response.
    :type challenge: str
    """

    def __init__(self, challenge):
        if not challenge:
            raise ValueError("Challenge cannot be empty")

        self._parameters = {}

        # Split the scheme ("Bearer") from the challenge parameters
        trimmed_challenge = challenge.strip()
        split_challenge = trimmed_challenge.split(" ", 1)
        trimmed_challenge = split_challenge[1]

        # Split trimmed challenge into name=value pairs; these pairs are expected to be split by either commas or spaces
        # Values may be surrounded by quotes, which are stripped here
        separator = "," if "," in trimmed_challenge else " "
        for item in trimmed_challenge.split(separator):
            # Process 'name=value' pairs
            comps = item.split("=")
            if len(comps) == 2:
                key = comps[0].strip(' "')
                value = comps[1].strip(' "')
                if key:
                    self._parameters[key] = value

        # Challenge must specify authorization or authorization_uri
        if not self._parameters or (
            "authorization" not in self._parameters and "authorization_uri" not in self._parameters
        ):
            raise ValueError("Invalid challenge parameters. `authorization` or `authorization_uri` must be present.")

        authorization_uri = self._parameters.get("authorization") or self._parameters.get("authorization_uri") or ""
        # the authorization server URI should look something like https://login.windows.net/tenant-id[/oauth2/authorize]
        uri_path = urlparse(authorization_uri).path.lstrip("/")
        self.tenant_id = uri_path.split("/")[0] or None

        self.scope = self._parameters.get("scope") or ""
        self.resource = self._parameters.get("resource") or self._parameters.get("resource_id") or ""


class SharedKeyCredentialPolicy(SansIOHTTPPolicy):
    def __init__(self, credential, is_emulated=False):
        self._credential = credential
        self.is_emulated = is_emulated

    def _get_headers(self, request, headers_to_sign):
        headers = dict((name.lower(), value) for name, value in request.headers.items() if value)
        if "content-length" in headers and headers["content-length"] == "0":
            del headers["content-length"]
        return "\n".join(headers.get(x, "") for x in headers_to_sign) + "\n"

    def _get_verb(self, request):
        return request.method + "\n"

    def _get_canonicalized_resource(self, request):
        uri_path = urlparse(request.http_request.url).path
        try:
            if (
                isinstance(request.context.transport, AsyncHttpTransport)
                or isinstance(
                    getattr(request.context.transport, "_transport", None),
                    AsyncHttpTransport,
                )
                or isinstance(
                    getattr(
                        getattr(request.context.transport, "_transport", None),
                        "_transport",
                        None,
                    ),
                    AsyncHttpTransport,
                )
            ):
                uri_path = URL(uri_path)
                return "/" + self._credential.named_key.name + str(uri_path)
        except TypeError:
            pass
        return "/" + self._credential.named_key.name + uri_path

    def _get_canonicalized_headers(self, request):
        string_to_sign = ""
        x_ms_headers = []
        for name, value in request.headers.items():
            if name.startswith("x-ms-"):
                x_ms_headers.append((name.lower(), value))
        x_ms_headers.sort()
        for name, value in x_ms_headers:
            if value is not None:
                string_to_sign += "".join([name, ":", value, "\n"])
        return string_to_sign

    def _add_authorization_header(self, request, string_to_sign):
        try:
            signature = _sign_string(self._credential.named_key.key, string_to_sign)
            auth_string = "SharedKey " + self._credential.named_key.name + ":" + signature
            request.headers["Authorization"] = auth_string
        except Exception as ex:
            # Wrap any error that occurred as signing error
            # Doing so will clarify/locate the source of problem
            raise _wrap_exception(ex, AzureSigningError) from ex

    def on_request(self, request: PipelineRequest) -> None:
        self.sign_request(request)

    def sign_request(self, request: PipelineRequest) -> None:
        string_to_sign = (
            self._get_verb(request.http_request)
            + self._get_headers(
                request.http_request,
                ["content-md5", "content-type", "x-ms-date"],
            )
            + self._get_canonicalized_resource(request)
            + self._get_canonicalized_resource_query(request.http_request)
        )
        self._add_authorization_header(request.http_request, string_to_sign)

    def _get_canonicalized_resource_query(self, request):
        for name, value in request.query.items():
            if name == "comp":
                return "?comp=" + value
        return ""


class BearerTokenChallengePolicy(BearerTokenCredentialPolicy):
    """Adds a bearer token Authorization header to requests, for the tenant provided in authentication challenges.

    See https://learn.microsoft.com/azure/active-directory/develop/claims-challenge for documentation on Microsoft
    Entra authentication challenges.

    :param credential: The credential.
    :type credential: ~azure.core.TokenCredential
    :param str scopes: Lets you specify the type of access needed.
    :keyword bool discover_tenant: Determines if tenant discovery should be enabled. Defaults to True.
    :keyword bool discover_scopes: Determines if scopes from authentication challenges should be provided to token
        requests, instead of the scopes given to the policy's constructor, if any are present. Defaults to True.
    :raises: :class:`~azure.core.exceptions.ServiceRequestError`
    """

    def __init__(
        self,
        credential: TokenCredential,
        *scopes: str,
        discover_tenant: bool = True,
        discover_scopes: bool = True,
        **kwargs,
    ) -> None:
        self._discover_tenant = discover_tenant
        self._discover_scopes = discover_scopes
        super().__init__(credential, *scopes, **kwargs)

    def on_challenge(self, request: PipelineRequest, response: PipelineResponse) -> bool:
        """Authorize request according to an authentication challenge

        This method is called when the resource provider responds 401 with a WWW-Authenticate header.

        :param ~azure.core.pipeline.PipelineRequest request: the request which elicited an authentication challenge
        :param ~azure.core.pipeline.PipelineResponse response: the resource provider's response
        :returns: a bool indicating whether the policy should send the request
        :rtype: bool
        """
        if not self._discover_tenant and not self._discover_scopes:
            # We can't discover the tenant or use a different scope; the request will fail because it hasn't changed
            return False

        try:
            challenge = _HttpChallenge(response.http_response.headers.get("WWW-Authenticate"))
            # azure-identity credentials require an AADv2 scope but the challenge may specify an AADv1 resource
            # if no scopes are included in the challenge, challenge.scope and challenge.resource will both be ''
            scope = challenge.scope or challenge.resource + "/.default" if self._discover_scopes else self._scopes
            if scope == "/.default":
                scope = self._scopes
        except ValueError:
            return False

        if self._discover_tenant:
            if isinstance(scope, str):
                self.authorize_request(request, scope, tenant_id=challenge.tenant_id)
            else:
                self.authorize_request(request, *scope, tenant_id=challenge.tenant_id)
        else:
            if isinstance(scope, str):
                self.authorize_request(request, scope)
            else:
                self.authorize_request(request, *scope)
        return True


@overload
def _configure_credential(
    credential: AzureNamedKeyCredential, cosmos_endpoint: bool = False, audience: Optional[str] = None
) -> SharedKeyCredentialPolicy: ...


@overload
def _configure_credential(
    credential: SharedKeyCredentialPolicy, cosmos_endpoint: bool = False, audience: Optional[str] = None
) -> SharedKeyCredentialPolicy: ...


@overload
def _configure_credential(
    credential: AzureSasCredential, cosmos_endpoint: bool = False, audience: Optional[str] = None
) -> AzureSasCredentialPolicy: ...


@overload
def _configure_credential(
    credential: TokenCredential, cosmos_endpoint: bool = False, audience: Optional[str] = None
) -> BearerTokenChallengePolicy: ...


@overload
def _configure_credential(credential: None, cosmos_endpoint: bool = False, audience: Optional[str] = None) -> None: ...


def _configure_credential(
    credential: Optional[
        Union[AzureNamedKeyCredential, AzureSasCredential, TokenCredential, SharedKeyCredentialPolicy]
    ],
    cosmos_endpoint: bool = False,
    audience: Optional[str] = None,
) -> Optional[Union[BearerTokenChallengePolicy, AzureSasCredentialPolicy, SharedKeyCredentialPolicy]]:
    if hasattr(credential, "get_token"):
        credential = cast(TokenCredential, credential)
        if audience:
            scope = audience.rstrip("/") + "/.default"
        else:
            scope = COSMOS_OAUTH_SCOPE if cosmos_endpoint else STORAGE_OAUTH_SCOPE
        return BearerTokenChallengePolicy(credential, scope)
    if isinstance(credential, SharedKeyCredentialPolicy):
        return credential
    if isinstance(credential, AzureSasCredential):
        return AzureSasCredentialPolicy(credential)
    if isinstance(credential, AzureNamedKeyCredential):
        return SharedKeyCredentialPolicy(credential)
    if credential is not None:
        raise TypeError(f"Unsupported credential: {type(credential)}")
    return None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from datetime import datetime, timedelta

import pytest

from devtools_testutils import AzureRecordedTestCase
from devtools_testutils.aio import recorded_by_proxy_async

from azure.core.credentials import AzureSasCredential
from azure.core.exceptions import ResourceExistsError
from azure.data.tables import (
    TableAccessPolicy,
    TableSasPermissions,
    ResourceTypes,
    AccountSasPermissions,
    TableItem,
    generate_account_sas,
)
from azure.data.tables.aio import TableServiceClient, TableClient

from _shared.asynctestcase import AsyncTableTestCase
from async_preparers import tables_decorator_async


class TestTableAsync(AzureRecordedTestCase, AsyncTableTestCase):
    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_create_table(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table_name = self._get_table_reference()

        # Act
        table = ts.get_table_client(table_name)
        created = await table.create_table()

        # Assert
        assert created.name == table_name
        await ts.delete_table(table_name=table_name)

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_create_table_fail_on_exist(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table_name = self._get_table_reference()

        # Act
        created = await ts.create_table(table_name=table_name)
        with pytest.raises(ResourceExistsError):
            await ts.create_table(table_name=table_name)

        name_filter = "TableName eq '{}'".format(table_name)
        existing = ts.query_tables(name_filter)

        # Assert
        assert isinstance(created, TableClient)
        await ts.delete_table(table_name=table_name)

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_query_tables_per_page(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table_name = "myasynctable"

        for i in range(5):
            await ts.create_table(table_name + str(i))

        query_filter = "TableName eq 'myasynctable0' or TableName eq 'myasynctable1' or TableName eq 'myasynctable2'"
        table_count = 0
        page_count = 0
        async for table_page in ts.query_tables(query_filter, results_per_page=2).by_page():

            temp_count = 0
            async for table in table_page:
                temp_count += 1
            assert temp_count <= 2
            page_count += 1
            table_count += temp_count

        assert page_count == 2
        assert table_count == 3

        for i in range(5):
            await ts.delete_table(table_name + str(i))

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_list_tables(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table = await self._create_table(ts)

        # Act
        tables = []
        async for t in ts.list_tables():
            tables.append(t)

        # Assert
        for table_item in tables:
            assert isinstance(table_item, TableItem)

        assert tables is not None
        assert len(tables) >= 1
        assert tables[0] is not None
        await ts.delete_table(table.table_name)

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_query_tables_with_filter(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table = await self._create_table(ts)

        # Act
        name_filter = "TableName eq '{}'".format(table.table_name)
        tables = []
        async for t in ts.query_tables(name_filter):
            tables.append(t)

        # Assert
        assert tables is not None
        assert len(tables) == 1
        for table_item in tables:
            assert isinstance(table_item, TableItem)
            assert table_item.name is not None
        await ts.delete_table(table.table_name)

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_list_tables_with_num_results(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        prefix = "listtable"
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        # Delete any existing tables
        async for table in ts.list_tables():
            await ts.delete_table(table.name)

        table_list = []
        for i in range(0, 4):
            await self._create_table(ts, prefix + str(i), table_list)

        # Act
        big_page = []
        async for t in ts.list_tables():
            big_page.append(t)

        small_page = []
        async for s in ts.list_tables(results_per_page=3).by_page():
            small_page.append(s)

        assert len(small_page) == 2
        assert len(big_page) >= 4

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_list_tables_with_marker(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        prefix = "listtable"
        table_names = []
        for i in range(0, 4):
            await self._create_table(ts, prefix + str(i), table_names)

        # Act
        generator1 = ts.list_tables(results_per_page=2).by_page()
        await generator1.__anext__()

        generator2 = ts.list_tables(results_per_page=2).by_page(continuation_token=generator1.continuation_token)
        await generator2.__anext__()

        tables1 = generator1._current_page
        tables2 = generator2._current_page

        tables1_len = 0
        async for _ in tables1:
            tables1_len += 1
        tables2_len = 0
        async for _ in tables2:
            tables2_len += 1

        # Assert
        assert tables1_len == 2
        assert tables2_len == 2
        assert tables1 != tables2

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_delete_table_with_existing_table(
        self, tables_storage_account_name, tables_primary_storage_account_key
    ):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table = await self._create_table(ts)

        # Act
        await ts.delete_table(table_name=table.table_name)

        existing = ts.query_tables("TableName eq '{}'".format(table.table_name))
        tables = []
        async for e in existing:
            tables.append(e)
        assert tables == []

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_delete_table_with_non_existing_table_fail_not_exist(
        self, tables_storage_account_name, tables_primary_storage_account_key
    ):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)
        table_name = self._get_table_reference()
        await ts.delete_table(table_name)

        # Assert

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_get_table_acl(self, tables_storage_account_name, tables_primary_storage_account_key):
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table = await self._create_table(ts)
        try:
            # Act
            acl = await table.get_table_access_policy()

            # Assert
            assert acl is not None
            assert len(acl) == 0
        finally:
            await ts.delete_table(table.table_name)

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_set_table_acl_with_empty_signed_identifiers(
        self, tables_storage_account_name, tables_primary_storage_account_key
    ):
        # Arrange
        account_url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table = await self._create_table(ts)
        try:
            # Act
            await table.set_table_access_policy(signed_identifiers={})

            # Assert
            acl = await table.get_table_access_policy()
            assert acl is not None
            assert len(acl) == 0
        finally:
            await ts.delete_table(table.table_name)

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_set_table_acl_with_empty_signed_identifier(
        self, tables_storage_account_name, tables_primary_storage_account_key
    ):
        # Arrange
        url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(url, credential=tables_primary_storage_account_key)
        table = await self._create_table(ts)
        try:
            dt = datetime(2021, 6, 8, 2, 10, 9)
            signed_identifiers = {
                "null": None,
                "empty": TableAccessPolicy(start=None, expiry=None, permission=None),
                "partial": TableAccessPolicy(permission="r"),
                "full": TableAccessPolicy(start=dt, expiry=dt, permission="r"),
            }
            await table.set_table_access_policy(signed_identifiers)
            acl = await table.get_table_access_policy()
            assert acl is not None
            assert len(acl) == 4
            assert acl["null"] is None
            assert acl["empty"] is None
            assert acl["partial"] is not None
            assert acl["partial"].permission == "r"
            assert acl["partial"].expiry is None
            assert acl["partial"].start is None
            assert acl["full"] is not None
            assert acl["full"].permission == "r"
            self._assert_policy_datetime(dt, acl["full"].expiry)
            self._assert_policy_datetime(dt, acl["full"].start)

            signed_identifiers.pop("empty")
            signed_identifiers["partial"] = None

            await table.set_table_access_policy(signed_identifiers)
            acl = await table.get_table_access_policy()
            assert acl is not None
            assert len(acl) == 3
            assert "empty" not in acl
            assert acl["null"] is None
            assert acl["partial"] is None
            assert acl["full"] is not None
            assert acl["full"].permission == "r"
            self._assert_policy_datetime(dt, acl["full"].expiry)
            self._assert_policy_datetime(dt, acl["full"].start)
        finally:
            await ts.delete_table(table.table_name)

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_set_table_acl_with_signed_identifiers(
        self, tables_storage_account_name, tables_primary_storage_account_key
    ):
        # Arrange
        url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(url, credential=tables_primary_storage_account_key)
        table = await self._create_table(ts)
        client = ts.get_table_client(table_name=table.table_name)

        # Act
        start = datetime(2021, 6, 8, 2, 10, 9) - timedelta(minutes=5)
        expiry = datetime(2021, 6, 8, 2, 10, 9) + timedelta(hours=1)
        identifiers = {
            "testid": TableAccessPolicy(start=start, expiry=expiry, permission=TableSasPermissions(read=True))
        }
        try:
            await client.set_table_access_policy(signed_identifiers=identifiers)

            # Assert
            acl = await client.get_table_access_policy()
            assert acl is not None
            assert len(acl) == 1
            assert acl.get("testid")
            self._assert_policy_datetime(start, acl["testid"].start)
            self._assert_policy_datetime(expiry, acl["testid"].expiry)
            assert acl["testid"].permission == "r"
        finally:
            await ts.delete_table(table.table_name)

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_set_table_acl_too_many_ids(self, tables_storage_account_name, tables_primary_storage_account_key):
        # Arrange
        url = self.account_url(tables_storage_account_name, "table")
        ts = TableServiceClient(url, credential=tables_primary_storage_account_key)
        table = await self._create_table(ts)
        try:
            # Act
            identifiers = {}
            for i in range(0, 6):
                identifiers["id{}".format(i)] = None

            # Assert
            with pytest.raises(ValueError):
                await table.set_table_access_policy(signed_identifiers=identifiers)
        finally:
            await ts.delete_table(table.table_name)

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_account_sas(self, tables_storage_account_name, tables_primary_storage_account_key):
        account_url = self.account_url(tables_storage_account_name, "table")
        tsc = TableServiceClient(credential=tables_primary_storage_account_key, endpoint=account_url)

        table = await self._create_table(tsc)
        try:
            entity = {
                "PartitionKey": "test",
                "RowKey": "test1",
                "text": "hello",
            }
            await table.upsert_entity(entity=entity)

            entity["RowKey"] = "test2"
            await table.upsert_entity(entity=entity)

            token = self.generate_sas(
                generate_account_sas,
                tables_primary_storage_account_key,
                resource_types=ResourceTypes(object=True),
                permission=AccountSasPermissions(read=True),
                expiry=datetime.utcnow() + timedelta(hours=1),
                start=datetime.utcnow() - timedelta(minutes=1),
            )

            token = AzureSasCredential(token)

            account_url = self.account_url(tables_storage_account_name, "table")
            service = TableServiceClient(credential=token, endpoint=account_url)

            # Act
            sas_table = service.get_table_client(table.table_name)
            entities = []
            async for e in sas_table.list_entities():
                entities.append(e)

            # Assert
            assert len(entities) == 2
            assert entities[0]["text"] == "hello"
            assert entities[1]["text"] == "hello"
        finally:
            await tsc.delete_table(table.table_name)

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_unicode_create_table_unicode_name(
        self, tables_storage_account_name, tables_primary_storage_account_key
    ):
        account_url = self.account_url(tables_storage_account_name, "table")
        tsc = TableServiceClient(account_url, credential=tables_primary_storage_account_key)
        invalid_table_name = "啊齄丂狛狜"

        with pytest.raises(ValueError) as excinfo:
            async with tsc:
                await tsc.create_table(invalid_table_name)
            assert (
                "Storage table names must be alphanumeric, cannot begin with a number, and must be between 3-63 characters long."
                in str(excinfo)
            )

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_create_table_invalid_name(self, tables_storage_account_name, tables_primary_storage_account_key):
        account_url = self.account_url(tables_storage_account_name, "table")
        tsc = TableServiceClient(account_url, credential=tables_primary_storage_account_key)
        invalid_table_name = "my_table"

        with pytest.raises(ValueError) as excinfo:
            async with tsc:
                await tsc.create_table(table_name=invalid_table_name)
            assert (
                "Storage table names must be alphanumeric, cannot begin with a number, and must be between 3-63 characters long."
                in str(excinfo)
            )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import uuid

from ._base import _TableTest, get_base_entity


class CreateEntityBatchTest(_TableTest):
    def __init__(self, arguments):
        super().__init__(arguments)
        self.base_entity = get_base_entity(self.args.full_edm)
        self.base_entity["PartitionKey"] = str(uuid.uuid4())

    def run_sync(self):
        batch_size = 0
        batch = []
        for row in range(self.args.count):
            self.base_entity["RowKey"] = str(row)
            batch.append(("upsert", self.base_entity))
            batch_size += 1
            if batch_size >= 100:
                self.table_client.submit_transaction(batch)
                batch = []
                batch_size = 0
        if batch_size:
            self.table_client.submit_transaction(batch)

    async def run_async(self):
        batch_size = 0
        batch = []
        for row in range(self.args.count):
            self.base_entity["RowKey"] = str(row)
            batch.append(("upsert", self.base_entity))
            batch_size += 1
            if batch_size >= 100:
                await self.async_table_client.submit_transaction(batch)
                batch = []
                batch_size = 0
        if batch_size:
            await self.async_table_client.submit_transaction(batch)

    @staticmethod
    def add_arguments(parser):
        super(CreateEntityBatchTest, CreateEntityBatchTest).add_arguments(parser)
        parser.add_argument(
            "-c",
            "--count",
            nargs="?",
            type=int,
            help="Number of entities to batch create. Defaults to 100",
            default=100,
        )



# coding: utf-8

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: sample_create_client_async.py

DESCRIPTION:
    These samples demonstrate creating a TableServiceClient and a TableClient

USAGE:
    python sample_create_client_async.py

    Set the environment variables with your own values before running the sample:
    1) TABLES_STORAGE_ENDPOINT_SUFFIX - the Table service account URL suffix
    2) TABLES_STORAGE_ACCOUNT_NAME - the name of the storage account
    3) TABLES_PRIMARY_STORAGE_ACCOUNT_KEY - the storage account access key
"""


import os
import asyncio
from dotenv import find_dotenv, load_dotenv


class CreateClients(object):
    def __init__(self):
        load_dotenv(find_dotenv())
        self.access_key = os.environ["TABLES_PRIMARY_STORAGE_ACCOUNT_KEY"]
        self.endpoint_suffix = os.environ["TABLES_STORAGE_ENDPOINT_SUFFIX"]
        self.account_name = os.environ["TABLES_STORAGE_ACCOUNT_NAME"]
        self.endpoint = f"{self.account_name}.table.{self.endpoint_suffix}"
        self.connection_string = f"DefaultEndpointsProtocol=https;AccountName={self.account_name};AccountKey={self.access_key};EndpointSuffix={self.endpoint_suffix}"

    async def create_table_client(self):
        # Instantiate a TableServiceClient using a connection string
        # [START create_table_client]
        from azure.data.tables.aio import TableClient

        async with TableClient.from_connection_string(
            conn_str=self.connection_string, table_name="tableName"
        ) as table_client:
            print(f"Table name: {table_client.table_name}")
        # [END create_table_client]

    async def create_table_service_client(self):
        # Instantiate a TableServiceClient using a shared access key
        # [START create_table_service_client]
        from azure.data.tables.aio import TableServiceClient
        from azure.core.credentials import AzureNamedKeyCredential

        credential = AzureNamedKeyCredential(self.account_name, self.access_key)
        async with TableServiceClient(endpoint=self.endpoint, credential=credential) as table_service:
            properties = await table_service.get_service_properties()
            print(f"Properties: {properties}")
        # [END create_table_service_client]


async def main():
    sample = CreateClients()
    await sample.create_table_client()
    await sample.create_table_service_client()


if __name__ == "__main__":
    asyncio.run(main())



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import uuid

from ._base import _TableTest, get_base_entity


class ListEntitiesTest(_TableTest):
    def __init__(self, arguments):
        super().__init__(arguments)
        self.partition_key = str(uuid.uuid4())

    async def global_setup(self):
        await super().global_setup()
        batch_size = 0
        batch = []
        for row in range(self.args.count):
            base_entity = get_base_entity(self.args.full_edm)
            base_entity["PartitionKey"] = self.partition_key
            base_entity["RowKey"] = str(row)
            batch.append(("create", base_entity))
            batch_size += 1
            if batch_size >= 100:
                await self.async_table_client.submit_transaction(batch)
                batch = []
                batch_size = 0
        if batch_size:
            await self.async_table_client.submit_transaction(batch)

    def run_sync(self):
        for _ in self.table_client.list_entities():
            pass

    async def run_async(self):
        async for _ in self.async_table_client.list_entities():
            pass

    @staticmethod
    def add_arguments(parser):
        super(ListEntitiesTest, ListEntitiesTest).add_arguments(parser)
        parser.add_argument(
            "-c", "--count", nargs="?", type=int, help="Number of entities to list. Defaults to 100", default=100
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from typing import Union, Optional, cast, overload

from azure.core.credentials import AzureSasCredential, AzureNamedKeyCredential
from azure.core.credentials_async import AsyncTokenCredential
from azure.core.pipeline import PipelineResponse, PipelineRequest
from azure.core.pipeline.policies import AsyncBearerTokenCredentialPolicy

from .._constants import STORAGE_OAUTH_SCOPE, COSMOS_OAUTH_SCOPE
from .._authentication import _HttpChallenge, AzureSasCredentialPolicy, SharedKeyCredentialPolicy


class AsyncBearerTokenChallengePolicy(AsyncBearerTokenCredentialPolicy):
    """Adds a bearer token Authorization header to requests, for the tenant provided in authentication challenges.

    See https://learn.microsoft.com/azure/active-directory/develop/claims-challenge for documentation on Microsoft
    Entra authentication challenges.

    :param credential: The credential.
    :type credential: ~azure.core.credentials_async.AsyncTokenCredential
    :param str scopes: Lets you specify the type of access needed.
    :keyword bool discover_tenant: Determines if tenant discovery should be enabled. Defaults to True.
    :keyword bool discover_scopes: Determines if scopes from authentication challenges should be provided to token
        requests, instead of the scopes given to the policy's constructor, if any are present. Defaults to True.
    """

    def __init__(
        self,
        credential: AsyncTokenCredential,
        *scopes: str,
        discover_tenant: bool = True,
        discover_scopes: bool = True,
        **kwargs,
    ) -> None:
        self._discover_tenant = discover_tenant
        self._discover_scopes = discover_scopes
        super().__init__(credential, *scopes, **kwargs)

    async def on_challenge(self, request: PipelineRequest, response: PipelineResponse) -> bool:
        """Authorize request according to an authentication challenge

        This method is called when the resource provider responds 401 with a WWW-Authenticate header.

        :param ~azure.core.pipeline.PipelineRequest request: the request which elicited an authentication challenge
        :param ~azure.core.pipeline.PipelineResponse response: the resource provider's response
        :returns: a bool indicating whether the policy should send the request
        :rtype: bool
        """
        if not self._discover_tenant and not self._discover_scopes:
            # We can't discover the tenant or use a different scope; the request will fail because it hasn't changed
            return False

        try:
            challenge = _HttpChallenge(response.http_response.headers.get("WWW-Authenticate"))
            # azure-identity credentials require an AADv2 scope but the challenge may specify an AADv1 resource
            # if no scopes are included in the challenge, challenge.scope and challenge.resource will both be ''
            scope = challenge.scope or challenge.resource + "/.default" if self._discover_scopes else self._scopes
            if scope == "/.default":
                scope = self._scopes
        except ValueError:
            return False

        if self._discover_tenant:
            if isinstance(scope, str):
                await self.authorize_request(request, scope, tenant_id=challenge.tenant_id)
            else:
                await self.authorize_request(request, *scope, tenant_id=challenge.tenant_id)
        else:
            if isinstance(scope, str):
                await self.authorize_request(request, scope)
            else:
                await self.authorize_request(request, *scope)
        return True


@overload
def _configure_credential(
    credential: AzureNamedKeyCredential, cosmos_endpoint: bool = False, audience: Optional[str] = None
) -> SharedKeyCredentialPolicy: ...


@overload
def _configure_credential(
    credential: SharedKeyCredentialPolicy, cosmos_endpoint: bool = False, audience: Optional[str] = None
) -> SharedKeyCredentialPolicy: ...


@overload
def _configure_credential(
    credential: AzureSasCredential, cosmos_endpoint: bool = False, audience: Optional[str] = None
) -> AzureSasCredentialPolicy: ...


@overload
def _configure_credential(
    credential: AsyncTokenCredential, cosmos_endpoint: bool = False, audience: Optional[str] = None
) -> AsyncBearerTokenChallengePolicy: ...


@overload
def _configure_credential(credential: None, cosmos_endpoint: bool = False, audience: Optional[str] = None) -> None: ...


def _configure_credential(
    credential: Optional[
        Union[AzureNamedKeyCredential, AzureSasCredential, AsyncTokenCredential, SharedKeyCredentialPolicy]
    ],
    cosmos_endpoint: bool = False,
    audience: Optional[str] = None,
) -> Optional[Union[AsyncBearerTokenChallengePolicy, AzureSasCredentialPolicy, SharedKeyCredentialPolicy]]:
    if hasattr(credential, "get_token"):
        credential = cast(AsyncTokenCredential, credential)
        if audience:
            scope = audience.rstrip("/") + "/.default"
        else:
            scope = COSMOS_OAUTH_SCOPE if cosmos_endpoint else STORAGE_OAUTH_SCOPE
        return AsyncBearerTokenChallengePolicy(credential, scope)
    if isinstance(credential, SharedKeyCredentialPolicy):
        return credential
    if isinstance(credential, AzureSasCredential):
        return AzureSasCredentialPolicy(credential)
    if isinstance(credential, AzureNamedKeyCredential):
        return SharedKeyCredentialPolicy(credential)
    if credential is not None:
        raise TypeError(f"Unsupported credential: {type(credential)}")
    return None



# coding: utf-8

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: sample_authentication.py

DESCRIPTION:
    These samples demonstrate authenticating a client via:
        * connection string
        * shared access key
        * generating a sas token with which the returned signature can be used with
    the credential parameter of any TableServiceClient or TableClient
        * Microsoft Entra ID

USAGE:
    python sample_authentication.py

    Set the environment variables with your own values before running the sample:
    1) TABLES_STORAGE_ENDPOINT_SUFFIX - the Table service account URL suffix
    2) TABLES_STORAGE_ACCOUNT_NAME - the name of the storage account
    3) TABLES_PRIMARY_STORAGE_ACCOUNT_KEY - the storage account access key
    The following environment variables are required for using azure-identity's DefaultAzureCredential.
    For more information, please refer to https://aka.ms/azsdk/python/identity/docs#azure.identity.DefaultAzureCredential
    4) AZURE_TENANT_ID - the tenant ID in Azure Active Directory
    5) AZURE_CLIENT_ID - the application (client) ID registered in the Microsoft Entra tenant
    6) AZURE_CLIENT_SECRET - the client secret for the registered application
"""

import os
from dotenv import find_dotenv, load_dotenv


class TableAuthSamples(object):
    def __init__(self):
        load_dotenv(find_dotenv())
        self.access_key = os.environ["TABLES_PRIMARY_STORAGE_ACCOUNT_KEY"]
        self.endpoint_suffix = os.environ["TABLES_STORAGE_ENDPOINT_SUFFIX"]
        self.account_name = os.environ["TABLES_STORAGE_ACCOUNT_NAME"]
        self.endpoint = f"{self.account_name}.table.{self.endpoint_suffix}"
        self.connection_string = f"DefaultEndpointsProtocol=https;AccountName={self.account_name};AccountKey={self.access_key};EndpointSuffix={self.endpoint_suffix}"

    def authentication_by_connection_string(self):
        print("Instantiate a TableServiceClient using a connection string")
        # [START auth_from_connection_string]
        from azure.data.tables import TableServiceClient

        with TableServiceClient.from_connection_string(conn_str=self.connection_string) as table_service_client:
            properties = table_service_client.get_service_properties()
            print(f"{properties}")
        # [END auth_from_connection_string]

    def authentication_by_shared_key(self):
        print("Instantiate a TableServiceClient using a shared access key")
        # [START auth_from_shared_key]
        from azure.data.tables import TableServiceClient
        from azure.core.credentials import AzureNamedKeyCredential

        credential = AzureNamedKeyCredential(self.account_name, self.access_key)
        with TableServiceClient(endpoint=self.endpoint, credential=credential) as table_service_client:
            properties = table_service_client.get_service_properties()
            print(f"{properties}")
        # [END auth_from_shared_key]

    def authentication_by_shared_access_signature(self):
        print("Instantiate a TableServiceClient using a shared access signature")
        # [START auth_from_sas]
        from datetime import datetime, timedelta
        from azure.data.tables import TableServiceClient, generate_account_sas, ResourceTypes, AccountSasPermissions
        from azure.core.credentials import AzureNamedKeyCredential, AzureSasCredential

        credential = AzureNamedKeyCredential(self.account_name, self.access_key)
        # Create a SAS token to use for authentication of a client
        sas_token = generate_account_sas(
            credential,
            resource_types=ResourceTypes(service=True),
            permission=AccountSasPermissions(read=True),
            expiry=datetime.utcnow() + timedelta(hours=1),
        )

        with TableServiceClient(
            endpoint=self.endpoint, credential=AzureSasCredential(sas_token)
        ) as table_service_client:
            properties = table_service_client.get_service_properties()
            print(f"{properties}")
        # [END auth_from_sas]

    def authentication_by_entra(self):
        print("Instantiate a TableServiceClient using a TokenCredential")
        # [START auth_from_aad]
        from azure.data.tables import TableServiceClient
        from azure.identity import DefaultAzureCredential

        with TableServiceClient(endpoint=self.endpoint, credential=DefaultAzureCredential()) as table_service_client:
            properties = table_service_client.get_service_properties()
            print(f"{properties}")
        # [END auth_from_aad]


if __name__ == "__main__":
    sample = TableAuthSamples()
    sample.authentication_by_connection_string()
    sample.authentication_by_shared_key()
    sample.authentication_by_shared_access_signature()
    sample.authentication_by_entra()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------

from unittest.mock import Mock

from azure.core.credentials import AccessToken
from azure.core.exceptions import ClientAuthenticationError
from azure.core.pipeline import AsyncPipeline
from azure.core.pipeline.transport import HttpRequest as PipelineTransportHttpRequest
from azure.data.tables.aio import TableServiceClient
from azure.data.tables.aio._authentication_async import AsyncBearerTokenChallengePolicy
import pytest

from async_preparers import tables_decorator_async
from devtools_testutils import AzureRecordedTestCase, is_live
from devtools_testutils.aio import recorded_by_proxy_async
from _shared.asynctestcase import AsyncTableTestCase

pytestmark = pytest.mark.asyncio

# Try to use azure.core.rest request if azure-core version supports it -- fall back to basic request
try:
    from azure.core.rest import HttpRequest as RestHttpRequest

    HTTP_REQUESTS = [PipelineTransportHttpRequest, RestHttpRequest]
except ModuleNotFoundError:
    HTTP_REQUESTS = [PipelineTransportHttpRequest]


class TestTableChallengeAuthAsync(AzureRecordedTestCase, AsyncTableTestCase):
    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_challenge_auth_supported_version(self, tables_storage_account_name):
        """This test requires a client that uses an API version that supports 401 challenge responses.

        Recorded using an incorrect tenant for the credential provided to our client. To run this live, ensure that the
        service principal used for testing is enabled for multitenant authentication
        (https://learn.microsoft.com/azure/active-directory/develop/howto-convert-app-to-be-multi-tenant). Set the
        TABLES_TENANT_ID environment variable to a different, existing tenant than the one the storage account exists
        in, and set CHALLENGE_TABLES_TENANT_ID to the tenant that the storage account exists in.
        """

        if is_live():
            pytest.skip("Playback testing and live manual testing only")

        account_url = self.account_url(tables_storage_account_name, "table")
        client = TableServiceClient(
            credential=self.get_token_credential(), endpoint=account_url, api_version="2020-12-06"
        )
        stats = await client.get_service_stats()
        self._assert_stats_default(stats)

    @tables_decorator_async
    @recorded_by_proxy_async
    async def test_challenge_auth_unsupported_version(self, tables_storage_account_name):
        """This test requires a client that uses an API version that doesn't support 401 challenge responses.

        Recorded using an incorrect tenant for the credential provided to our client. To run this live, ensure that the
        service principal used for testing is enabled for multitenant authentication
        (https://learn.microsoft.com/azure/active-directory/develop/howto-convert-app-to-be-multi-tenant). Set the
        TABLES_TENANT_ID environment variable to a different, existing tenant than the one the storage account exists
        in, and set CHALLENGE_TABLES_TENANT_ID to the tenant that the storage account exists in.
        """

        if is_live():
            pytest.skip("Playback testing and live manual testing only")

        account_url = self.account_url(tables_storage_account_name, "table")
        client = TableServiceClient(
            credential=self.get_token_credential(), endpoint=account_url, api_version="2019-07-07"
        )

        # Our request fails because of the invalid token, prompting a 403 response
        with pytest.raises(ClientAuthenticationError) as e:
            await client.get_service_stats()


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_challenge_policy_uses_scopes_and_tenant(http_request):
    """The policy's token requests should pass the parsed scope and tenant ID from the challenge"""

    async def test_with_challenge(challenge, expected_scope, expected_tenant):
        expected_token = "expected_token"

        class Requests:
            count = 0

        class TokenRequests:
            count = 0

        async def send(request):
            Requests.count += 1
            if Requests.count == 1:
                # first request triggers a 401 response w/ auth challenge
                return challenge
            elif Requests.count == 2:
                # second request should be authorized according to challenge and have the expected content
                assert expected_token in request.headers["Authorization"]
                return Mock(status_code=200)
            raise ValueError("unexpected request")

        async def get_token(*scopes, **kwargs):
            assert len(scopes) == 1
            TokenRequests.count += 1
            if TokenRequests.count == 1:
                # first request uses the scope given to the policy, and no tenant ID
                assert scopes[0] != expected_scope
                assert kwargs.get("tenant_id") is None
            elif TokenRequests.count == 2:
                # second request should use the scope and tenant ID specified in the auth challenge
                assert scopes[0] == expected_scope
                assert kwargs.get("tenant_id") == expected_tenant
            else:
                raise ValueError("unexpected token request")
            return AccessToken(expected_token, 0)

        credential = Mock(spec_set=["get_token"], get_token=Mock(wraps=get_token))
        policy = AsyncBearerTokenChallengePolicy(credential, "scope")
        pipeline = AsyncPipeline(policies=[policy], transport=Mock(send=send))
        await pipeline.run(http_request("GET", "https://localhost"))

    tenant = "tenant-id"
    endpoint = f"https://authority.net/{tenant}/oauth2/authorize"
    resource = "https://challenge.resource"
    scope = f"{resource}/.default"

    # this challenge separates the authorization server and resource with commas in the WWW-Authenticate header
    challenge_with_commas = Mock(
        status_code=401,
        headers={"WWW-Authenticate": f'Bearer authorization="{endpoint}", resource="{resource}"'},
    )
    await test_with_challenge(challenge_with_commas, scope, tenant)

    # this challenge separates the authorization server and resource with only spaces in the WWW-Authenticate header
    challenge_without_commas = Mock(
        status_code=401,
        headers={"WWW-Authenticate": f"Bearer authorization={endpoint} resource={resource}"},
    )
    await test_with_challenge(challenge_without_commas, scope, tenant)

    # this challenge gives an AADv2 scope, ending with "/.default", instead of an AADv1 resource
    challenge_with_scope = Mock(
        status_code=401,
        headers={"WWW-Authenticate": f"Bearer authorization={endpoint} scope={scope}"},
    )
    await test_with_challenge(challenge_with_scope, scope, tenant)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_challenge_policy_disable_tenant_discovery(http_request):
    """The policy's token requests should exclude the challenge's tenant if requested"""

    async def test_with_challenge(challenge, challenge_scope):
        bad_token = "bad_token"

        class Requests:
            count = 0

        class TokenRequests:
            count = 0

        async def send(request):
            Requests.count += 1
            if Requests.count == 1:
                # first request triggers a 401 response w/ auth challenge
                assert bad_token in request.headers["Authorization"]
                return challenge
            elif Requests.count == 2:
                # second request should still be unauthorized because we didn't authenticate in the correct tenant
                assert bad_token in request.headers["Authorization"]
                return challenge
            raise ValueError("unexpected request")

        async def get_token(*scopes, **kwargs):
            assert len(scopes) == 1
            TokenRequests.count += 1
            if TokenRequests.count == 1:
                # first request uses the scope given to the policy, and no tenant ID
                assert scopes[0] != challenge_scope
                assert kwargs.get("tenant_id") is None
                return AccessToken(bad_token, 0)
            elif TokenRequests.count == 2:
                # second request should use the scope specified in the auth challenge, but not the tenant
                assert scopes[0] == challenge_scope
                assert kwargs.get("tenant_id") is None
                return AccessToken(bad_token, 0)
            raise ValueError("unexpected token request")

        credential = Mock(spec_set=["get_token"], get_token=Mock(wraps=get_token))
        policy = AsyncBearerTokenChallengePolicy(credential, "scope", discover_tenant=False)
        pipeline = AsyncPipeline(policies=[policy], transport=Mock(send=send))
        await pipeline.run(http_request("GET", "https://localhost"))

    tenant = "tenant-id"
    endpoint = f"https://authority.net/{tenant}/oauth2/authorize"
    resource = "https://challenge.resource"
    scope = f"{resource}/.default"

    # this challenge separates the authorization server and resource with commas in the WWW-Authenticate header
    challenge_with_commas = Mock(
        status_code=401,
        headers={"WWW-Authenticate": f'Bearer authorization="{endpoint}", resource="{resource}"'},
    )
    # the request should fail after the challenge because we don't use the correct tenant
    # after the second 4xx response, the policy should raise the authentication error
    await test_with_challenge(challenge_with_commas, scope)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_challenge_policy_disable_scopes_discovery(http_request):
    """The policy's token requests should exclude the challenge's scope if requested"""

    async def test_with_challenge(challenge, challenge_scope, challenge_tenant):
        bad_token = "bad_token"

        class Requests:
            count = 0

        class TokenRequests:
            count = 0

        async def send(request):
            Requests.count += 1
            if Requests.count == 1:
                # first request triggers a 401 response w/ auth challenge
                assert bad_token in request.headers["Authorization"]
                return challenge
            elif Requests.count == 2:
                # second request should still be unauthorized because we didn't authenticate with the correct scope
                assert bad_token in request.headers["Authorization"]
                return challenge
            raise ValueError("unexpected request")

        async def get_token(*scopes, **kwargs):
            assert len(scopes) == 1
            TokenRequests.count += 1
            if TokenRequests.count == 1:
                # first request uses the scope given to the policy, and no tenant ID
                assert scopes[0] != challenge_scope
                assert kwargs.get("tenant_id") is None
                return AccessToken(bad_token, 0)
            elif TokenRequests.count == 2:
                # second request should use the tenant specified in the auth challenge, but not the scope
                assert scopes[0] != challenge_scope
                assert kwargs.get("tenant_id") == challenge_tenant
                return AccessToken(bad_token, 0)
            raise ValueError("unexpected token request")

        credential = Mock(spec_set=["get_token"], get_token=Mock(wraps=get_token))
        policy = AsyncBearerTokenChallengePolicy(credential, ["scope1", "scope2"], discover_scopes=False)
        pipeline = AsyncPipeline(policies=[policy], transport=Mock(send=send))
        await pipeline.run(http_request("GET", "https://localhost"))

    tenant = "tenant-id"
    endpoint = f"https://authority.net/{tenant}/oauth2/authorize"
    resource = "https://challenge.resource"
    scope = f"{resource}/.default"

    # this challenge separates the authorization server and resource with commas in the WWW-Authenticate header
    challenge_with_commas = Mock(
        status_code=401,
        headers={"WWW-Authenticate": f'Bearer authorization="{endpoint}", resource="{resource}"'},
    )
    # the request should fail after the challenge because we don't use the correct scope
    # after the second 4xx response, the policy should raise the authentication error
    await test_with_challenge(challenge_with_commas, scope, tenant)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_challenge_policy_disable_any_discovery(http_request):
    """The policy shouldn't respond to the challenge if it can't use the provided scope or tenant"""

    async def test_with_challenge(challenge, challenge_scope, challenge_tenant):
        bad_token = "bad_token"

        class Requests:
            count = 0

        class TokenRequests:
            count = 0

        async def send(request):
            Requests.count += 1
            if Requests.count == 1:
                # first request triggers a 401 response w/ auth challenge
                assert bad_token in request.headers["Authorization"]
                return challenge
            raise ValueError("unexpected request")

        async def get_token(*scopes, **kwargs):
            assert len(scopes) == 1
            TokenRequests.count += 1
            if TokenRequests.count == 1:
                # first request uses the scope given to the policy, and no tenant ID
                assert scopes[0] != challenge_scope
                assert kwargs.get("tenant_id") is None
                return AccessToken(bad_token, 0)
            raise ValueError("unexpected token request")

        credential = Mock(spec_set=["get_token"], get_token=Mock(wraps=get_token))
        policy = AsyncBearerTokenChallengePolicy(
            credential, ["scope1", "scope2"], discover_tenant=False, discover_scopes=False
        )
        pipeline = AsyncPipeline(policies=[policy], transport=Mock(send=send))
        await pipeline.run(http_request("GET", "https://localhost"))

    tenant = "tenant-id"
    endpoint = f"https://authority.net/{tenant}/oauth2/authorize"
    resource = "https://challenge.resource"
    scope = f"{resource}/.default"

    # this challenge separates the authorization server and resource with commas in the WWW-Authenticate header
    challenge_with_commas = Mock(
        status_code=401,
        headers={"WWW-Authenticate": f'Bearer authorization="{endpoint}", resource="{resource}"'},
    )
    # the policy should only send one request since we can't update our request per the challenge response
    await test_with_challenge(challenge_with_commas, scope, tenant)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_challenge_policy_no_scope_in_challenge(http_request):
    """The policy's token requests should use constructor scopes if none are in the challenge"""

    async def test_with_challenge(challenge, challenge_scope, challenge_tenant):
        bad_token = "bad_token"

        class Requests:
            count = 0

        class TokenRequests:
            count = 0

        async def send(request):
            Requests.count += 1
            if Requests.count == 1:
                # first request triggers a 401 response w/ auth challenge
                assert bad_token in request.headers["Authorization"]
                return challenge
            elif Requests.count == 2:
                # second request should still be unauthorized because we didn't authenticate with the correct scope
                assert bad_token in request.headers["Authorization"]
                return challenge
            raise ValueError("unexpected request")

        async def get_token(*scopes, **kwargs):
            assert len(scopes) == 1
            TokenRequests.count += 1
            if TokenRequests.count == 1:
                # first request uses the scope given to the policy, and no tenant ID
                assert scopes[0] != challenge_scope
                assert kwargs.get("tenant_id") is None
                return AccessToken(bad_token, 0)
            elif TokenRequests.count == 2:
                # second request should use the tenant specified in the auth challenge, and same scope as before
                assert scopes[0] != challenge_scope
                assert kwargs.get("tenant_id") == challenge_tenant
                return AccessToken(bad_token, 0)
            raise ValueError("unexpected token request")

        credential = Mock(spec_set=["get_token"], get_token=Mock(wraps=get_token))
        policy = AsyncBearerTokenChallengePolicy(credential, "scope")
        pipeline = AsyncPipeline(policies=[policy], transport=Mock(send=send))
        await pipeline.run(http_request("GET", "https://localhost"))

    tenant = "tenant-id"
    endpoint = f"https://authority.net/{tenant}/oauth2/authorize"
    resource = "https://challenge.resource"
    scope = f"{resource}/.default"

    # this challenge has no `resource` or `scope` field
    challenge_with_commas = Mock(
        status_code=401,
        headers={"WWW-Authenticate": f'Bearer authorization="{endpoint}"'},
    )
    # the request should fail after the challenge because we don't use the correct scope
    # after the second 4xx response, the policy should raise the authentication error
    await test_with_challenge(challenge_with_commas, scope, tenant)



# coding=utf-8
# pylint: disable=too-many-lines
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

import datetime
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

from .. import _serialization

if TYPE_CHECKING:
    # pylint: disable=unused-import,ungrouped-imports
    from .. import models as _models


class AccessPolicy(_serialization.Model):
    """An Access policy.

    All required parameters must be populated in order to send to Azure.

    :ivar start: The start datetime from which the policy is active. Required.
    :vartype start: str
    :ivar expiry: The datetime that the policy expires. Required.
    :vartype expiry: str
    :ivar permission: The permissions for the acl policy. Required.
    :vartype permission: str
    """

    _validation = {
        "start": {"required": True},
        "expiry": {"required": True},
        "permission": {"required": True},
    }

    _attribute_map = {
        "start": {"key": "Start", "type": "str", "xml": {"name": "Start"}},
        "expiry": {"key": "Expiry", "type": "str", "xml": {"name": "Expiry"}},
        "permission": {"key": "Permission", "type": "str", "xml": {"name": "Permission"}},
    }
    _xml_map = {"name": "AccessPolicy"}

    def __init__(self, *, start: str, expiry: str, permission: str, **kwargs: Any) -> None:
        """
        :keyword start: The start datetime from which the policy is active. Required.
        :paramtype start: str
        :keyword expiry: The datetime that the policy expires. Required.
        :paramtype expiry: str
        :keyword permission: The permissions for the acl policy. Required.
        :paramtype permission: str
        """
        super().__init__(**kwargs)
        self.start = start
        self.expiry = expiry
        self.permission = permission


class CorsRule(_serialization.Model):
    """CORS is an HTTP feature that enables a web application running under one domain to access
    resources in another domain. Web browsers implement a security restriction known as same-origin
    policy that prevents a web page from calling APIs in a different domain; CORS provides a secure
    way to allow one domain (the origin domain) to call APIs in another domain.

    All required parameters must be populated in order to send to Azure.

    :ivar allowed_origins: The origin domains that are permitted to make a request against the
     service via CORS. The origin domain is the domain from which the request originates. Note that
     the origin must be an exact case-sensitive match with the origin that the user age sends to the
     service. You can also use the wildcard character '*' to allow all origin domains to make
     requests via CORS. Required.
    :vartype allowed_origins: str
    :ivar allowed_methods: The methods (HTTP request verbs) that the origin domain may use for a
     CORS request. (comma separated). Required.
    :vartype allowed_methods: str
    :ivar allowed_headers: The request headers that the origin domain may specify on the CORS
     request. Required.
    :vartype allowed_headers: str
    :ivar exposed_headers: The response headers that may be sent in the response to the CORS
     request and exposed by the browser to the request issuer. Required.
    :vartype exposed_headers: str
    :ivar max_age_in_seconds: The maximum amount time that a browser should cache the preflight
     OPTIONS request. Required.
    :vartype max_age_in_seconds: int
    """

    _validation = {
        "allowed_origins": {"required": True},
        "allowed_methods": {"required": True},
        "allowed_headers": {"required": True},
        "exposed_headers": {"required": True},
        "max_age_in_seconds": {"required": True, "minimum": 0},
    }

    _attribute_map = {
        "allowed_origins": {"key": "AllowedOrigins", "type": "str", "xml": {"name": "AllowedOrigins"}},
        "allowed_methods": {"key": "AllowedMethods", "type": "str", "xml": {"name": "AllowedMethods"}},
        "allowed_headers": {"key": "AllowedHeaders", "type": "str", "xml": {"name": "AllowedHeaders"}},
        "exposed_headers": {"key": "ExposedHeaders", "type": "str", "xml": {"name": "ExposedHeaders"}},
        "max_age_in_seconds": {"key": "MaxAgeInSeconds", "type": "int", "xml": {"name": "MaxAgeInSeconds"}},
    }
    _xml_map = {"name": "CorsRule"}

    def __init__(
        self,
        *,
        allowed_origins: str,
        allowed_methods: str,
        allowed_headers: str,
        exposed_headers: str,
        max_age_in_seconds: int,
        **kwargs: Any
    ) -> None:
        """
        :keyword allowed_origins: The origin domains that are permitted to make a request against the
         service via CORS. The origin domain is the domain from which the request originates. Note that
         the origin must be an exact case-sensitive match with the origin that the user age sends to the
         service. You can also use the wildcard character '*' to allow all origin domains to make
         requests via CORS. Required.
        :paramtype allowed_origins: str
        :keyword allowed_methods: The methods (HTTP request verbs) that the origin domain may use for a
         CORS request. (comma separated). Required.
        :paramtype allowed_methods: str
        :keyword allowed_headers: The request headers that the origin domain may specify on the CORS
         request. Required.
        :paramtype allowed_headers: str
        :keyword exposed_headers: The response headers that may be sent in the response to the CORS
         request and exposed by the browser to the request issuer. Required.
        :paramtype exposed_headers: str
        :keyword max_age_in_seconds: The maximum amount time that a browser should cache the preflight
         OPTIONS request. Required.
        :paramtype max_age_in_seconds: int
        """
        super().__init__(**kwargs)
        self.allowed_origins = allowed_origins
        self.allowed_methods = allowed_methods
        self.allowed_headers = allowed_headers
        self.exposed_headers = exposed_headers
        self.max_age_in_seconds = max_age_in_seconds


class GeoReplication(_serialization.Model):
    """GeoReplication.

    All required parameters must be populated in order to send to Azure.

    :ivar status: The status of the secondary location. Required. Known values are: "live",
     "bootstrap", and "unavailable".
    :vartype status: str or ~azure.table.models.GeoReplicationStatusType
    :ivar last_sync_time: A GMT date/time value, to the second. All primary writes preceding this
     value are guaranteed to be available for read operations at the secondary. Primary writes after
     this point in time may or may not be available for reads. Required.
    :vartype last_sync_time: ~datetime.datetime
    """

    _validation = {
        "status": {"required": True},
        "last_sync_time": {"required": True},
    }

    _attribute_map = {
        "status": {"key": "Status", "type": "str", "xml": {"name": "Status"}},
        "last_sync_time": {"key": "LastSyncTime", "type": "rfc-1123", "xml": {"name": "LastSyncTime"}},
    }
    _xml_map = {"name": "GeoReplication"}

    def __init__(
        self,
        *,
        status: Union[str, "_models.GeoReplicationStatusType"],
        last_sync_time: datetime.datetime,
        **kwargs: Any
    ) -> None:
        """
        :keyword status: The status of the secondary location. Required. Known values are: "live",
         "bootstrap", and "unavailable".
        :paramtype status: str or ~azure.table.models.GeoReplicationStatusType
        :keyword last_sync_time: A GMT date/time value, to the second. All primary writes preceding
         this value are guaranteed to be available for read operations at the secondary. Primary writes
         after this point in time may or may not be available for reads. Required.
        :paramtype last_sync_time: ~datetime.datetime
        """
        super().__init__(**kwargs)
        self.status = status
        self.last_sync_time = last_sync_time


class Logging(_serialization.Model):
    """Azure Analytics Logging settings.

    All required parameters must be populated in order to send to Azure.

    :ivar version: The version of Analytics to configure. Required.
    :vartype version: str
    :ivar delete: Indicates whether all delete requests should be logged. Required.
    :vartype delete: bool
    :ivar read: Indicates whether all read requests should be logged. Required.
    :vartype read: bool
    :ivar write: Indicates whether all write requests should be logged. Required.
    :vartype write: bool
    :ivar retention_policy: The retention policy. Required.
    :vartype retention_policy: ~azure.table.models.RetentionPolicy
    """

    _validation = {
        "version": {"required": True},
        "delete": {"required": True},
        "read": {"required": True},
        "write": {"required": True},
        "retention_policy": {"required": True},
    }

    _attribute_map = {
        "version": {"key": "Version", "type": "str", "xml": {"name": "Version"}},
        "delete": {"key": "Delete", "type": "bool", "xml": {"name": "Delete"}},
        "read": {"key": "Read", "type": "bool", "xml": {"name": "Read"}},
        "write": {"key": "Write", "type": "bool", "xml": {"name": "Write"}},
        "retention_policy": {"key": "RetentionPolicy", "type": "RetentionPolicy"},
    }
    _xml_map = {"name": "Logging"}

    def __init__(
        self,
        *,
        version: str,
        delete: bool,
        read: bool,
        write: bool,
        retention_policy: "_models.RetentionPolicy",
        **kwargs: Any
    ) -> None:
        """
        :keyword version: The version of Analytics to configure. Required.
        :paramtype version: str
        :keyword delete: Indicates whether all delete requests should be logged. Required.
        :paramtype delete: bool
        :keyword read: Indicates whether all read requests should be logged. Required.
        :paramtype read: bool
        :keyword write: Indicates whether all write requests should be logged. Required.
        :paramtype write: bool
        :keyword retention_policy: The retention policy. Required.
        :paramtype retention_policy: ~azure.table.models.RetentionPolicy
        """
        super().__init__(**kwargs)
        self.version = version
        self.delete = delete
        self.read = read
        self.write = write
        self.retention_policy = retention_policy


class Metrics(_serialization.Model):
    """Metrics.

    All required parameters must be populated in order to send to Azure.

    :ivar version: The version of Analytics to configure.
    :vartype version: str
    :ivar enabled: Indicates whether metrics are enabled for the Table service. Required.
    :vartype enabled: bool
    :ivar include_apis: Indicates whether metrics should generate summary statistics for called API
     operations.
    :vartype include_apis: bool
    :ivar retention_policy: The retention policy.
    :vartype retention_policy: ~azure.table.models.RetentionPolicy
    """

    _validation = {
        "enabled": {"required": True},
    }

    _attribute_map = {
        "version": {"key": "Version", "type": "str", "xml": {"name": "Version"}},
        "enabled": {"key": "Enabled", "type": "bool", "xml": {"name": "Enabled"}},
        "include_apis": {"key": "IncludeAPIs", "type": "bool", "xml": {"name": "IncludeAPIs"}},
        "retention_policy": {"key": "RetentionPolicy", "type": "RetentionPolicy"},
    }

    def __init__(
        self,
        *,
        enabled: bool,
        version: Optional[str] = None,
        include_apis: Optional[bool] = None,
        retention_policy: Optional["_models.RetentionPolicy"] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword version: The version of Analytics to configure.
        :paramtype version: str
        :keyword enabled: Indicates whether metrics are enabled for the Table service. Required.
        :paramtype enabled: bool
        :keyword include_apis: Indicates whether metrics should generate summary statistics for called
         API operations.
        :paramtype include_apis: bool
        :keyword retention_policy: The retention policy.
        :paramtype retention_policy: ~azure.table.models.RetentionPolicy
        """
        super().__init__(**kwargs)
        self.version = version
        self.enabled = enabled
        self.include_apis = include_apis
        self.retention_policy = retention_policy


class RetentionPolicy(_serialization.Model):
    """The retention policy.

    All required parameters must be populated in order to send to Azure.

    :ivar enabled: Indicates whether a retention policy is enabled for the service. Required.
    :vartype enabled: bool
    :ivar days: Indicates the number of days that metrics or logging or soft-deleted data should be
     retained. All data older than this value will be deleted.
    :vartype days: int
    """

    _validation = {
        "enabled": {"required": True},
        "days": {"minimum": 1},
    }

    _attribute_map = {
        "enabled": {"key": "Enabled", "type": "bool", "xml": {"name": "Enabled"}},
        "days": {"key": "Days", "type": "int", "xml": {"name": "Days"}},
    }
    _xml_map = {"name": "RetentionPolicy"}

    def __init__(self, *, enabled: bool, days: Optional[int] = None, **kwargs: Any) -> None:
        """
        :keyword enabled: Indicates whether a retention policy is enabled for the service. Required.
        :paramtype enabled: bool
        :keyword days: Indicates the number of days that metrics or logging or soft-deleted data should
         be retained. All data older than this value will be deleted.
        :paramtype days: int
        """
        super().__init__(**kwargs)
        self.enabled = enabled
        self.days = days


class SignedIdentifier(_serialization.Model):
    """A signed identifier.

    All required parameters must be populated in order to send to Azure.

    :ivar id: A unique id. Required.
    :vartype id: str
    :ivar access_policy: The access policy.
    :vartype access_policy: ~azure.table.models.AccessPolicy
    """

    _validation = {
        "id": {"required": True},
    }

    _attribute_map = {
        "id": {"key": "Id", "type": "str", "xml": {"name": "Id"}},
        "access_policy": {"key": "AccessPolicy", "type": "AccessPolicy"},
    }
    _xml_map = {"name": "SignedIdentifier"}

    def __init__(
        self,
        *,
        id: str,  # pylint: disable=redefined-builtin
        access_policy: Optional["_models.AccessPolicy"] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword id: A unique id. Required.
        :paramtype id: str
        :keyword access_policy: The access policy.
        :paramtype access_policy: ~azure.table.models.AccessPolicy
        """
        super().__init__(**kwargs)
        self.id = id
        self.access_policy = access_policy


class TableEntityQueryResponse(_serialization.Model):
    """The properties for the table entity query response.

    :ivar odata_metadata: The metadata response of the table.
    :vartype odata_metadata: str
    :ivar value: List of table entities.
    :vartype value: list[dict[str, any]]
    """

    _attribute_map = {
        "odata_metadata": {"key": "odata\\.metadata", "type": "str"},
        "value": {"key": "value", "type": "[{object}]"},
    }

    def __init__(
        self, *, odata_metadata: Optional[str] = None, value: Optional[List[Dict[str, Any]]] = None, **kwargs: Any
    ) -> None:
        """
        :keyword odata_metadata: The metadata response of the table.
        :paramtype odata_metadata: str
        :keyword value: List of table entities.
        :paramtype value: list[dict[str, any]]
        """
        super().__init__(**kwargs)
        self.odata_metadata = odata_metadata
        self.value = value


class TableProperties(_serialization.Model):
    """The properties for creating a table.

    :ivar table_name: The name of the table to create.
    :vartype table_name: str
    """

    _attribute_map = {
        "table_name": {"key": "TableName", "type": "str"},
    }

    def __init__(self, *, table_name: Optional[str] = None, **kwargs: Any) -> None:
        """
        :keyword table_name: The name of the table to create.
        :paramtype table_name: str
        """
        super().__init__(**kwargs)
        self.table_name = table_name


class TableQueryResponse(_serialization.Model):
    """The properties for the table query response.

    :ivar odata_metadata: The metadata response of the table.
    :vartype odata_metadata: str
    :ivar value: List of tables.
    :vartype value: list[~azure.table.models.TableResponseProperties]
    """

    _attribute_map = {
        "odata_metadata": {"key": "odata\\.metadata", "type": "str"},
        "value": {"key": "value", "type": "[TableResponseProperties]"},
    }

    def __init__(
        self,
        *,
        odata_metadata: Optional[str] = None,
        value: Optional[List["_models.TableResponseProperties"]] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword odata_metadata: The metadata response of the table.
        :paramtype odata_metadata: str
        :keyword value: List of tables.
        :paramtype value: list[~azure.table.models.TableResponseProperties]
        """
        super().__init__(**kwargs)
        self.odata_metadata = odata_metadata
        self.value = value


class TableResponseProperties(_serialization.Model):
    """The properties for the table response.

    :ivar table_name: The name of the table.
    :vartype table_name: str
    :ivar odata_type: The odata type of the table.
    :vartype odata_type: str
    :ivar odata_id: The id of the table.
    :vartype odata_id: str
    :ivar odata_edit_link: The edit link of the table.
    :vartype odata_edit_link: str
    """

    _attribute_map = {
        "table_name": {"key": "TableName", "type": "str"},
        "odata_type": {"key": "odata\\.type", "type": "str"},
        "odata_id": {"key": "odata\\.id", "type": "str"},
        "odata_edit_link": {"key": "odata\\.editLink", "type": "str"},
    }

    def __init__(
        self,
        *,
        table_name: Optional[str] = None,
        odata_type: Optional[str] = None,
        odata_id: Optional[str] = None,
        odata_edit_link: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword table_name: The name of the table.
        :paramtype table_name: str
        :keyword odata_type: The odata type of the table.
        :paramtype odata_type: str
        :keyword odata_id: The id of the table.
        :paramtype odata_id: str
        :keyword odata_edit_link: The edit link of the table.
        :paramtype odata_edit_link: str
        """
        super().__init__(**kwargs)
        self.table_name = table_name
        self.odata_type = odata_type
        self.odata_id = odata_id
        self.odata_edit_link = odata_edit_link


class TableResponse(TableResponseProperties):
    """The response for a single table.

    :ivar table_name: The name of the table.
    :vartype table_name: str
    :ivar odata_type: The odata type of the table.
    :vartype odata_type: str
    :ivar odata_id: The id of the table.
    :vartype odata_id: str
    :ivar odata_edit_link: The edit link of the table.
    :vartype odata_edit_link: str
    :ivar odata_metadata: The metadata response of the table.
    :vartype odata_metadata: str
    """

    _attribute_map = {
        "table_name": {"key": "TableName", "type": "str"},
        "odata_type": {"key": "odata\\.type", "type": "str"},
        "odata_id": {"key": "odata\\.id", "type": "str"},
        "odata_edit_link": {"key": "odata\\.editLink", "type": "str"},
        "odata_metadata": {"key": "odata\\.metadata", "type": "str"},
    }

    def __init__(
        self,
        *,
        table_name: Optional[str] = None,
        odata_type: Optional[str] = None,
        odata_id: Optional[str] = None,
        odata_edit_link: Optional[str] = None,
        odata_metadata: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword table_name: The name of the table.
        :paramtype table_name: str
        :keyword odata_type: The odata type of the table.
        :paramtype odata_type: str
        :keyword odata_id: The id of the table.
        :paramtype odata_id: str
        :keyword odata_edit_link: The edit link of the table.
        :paramtype odata_edit_link: str
        :keyword odata_metadata: The metadata response of the table.
        :paramtype odata_metadata: str
        """
        super().__init__(
            table_name=table_name, odata_type=odata_type, odata_id=odata_id, odata_edit_link=odata_edit_link, **kwargs
        )
        self.odata_metadata = odata_metadata


class TableServiceError(_serialization.Model):
    """Table Service error.

    :ivar message: The error message.
    :vartype message: str
    """

    _attribute_map = {
        "message": {"key": "Message", "type": "str", "xml": {"name": "Message"}},
    }

    def __init__(self, *, message: Optional[str] = None, **kwargs: Any) -> None:
        """
        :keyword message: The error message.
        :paramtype message: str
        """
        super().__init__(**kwargs)
        self.message = message


class TableServiceProperties(_serialization.Model):
    """Table Service Properties.

    :ivar logging: Azure Analytics Logging settings.
    :vartype logging: ~azure.table.models.Logging
    :ivar hour_metrics: A summary of request statistics grouped by API in hourly aggregates for
     tables.
    :vartype hour_metrics: ~azure.table.models.Metrics
    :ivar minute_metrics: A summary of request statistics grouped by API in minute aggregates for
     tables.
    :vartype minute_metrics: ~azure.table.models.Metrics
    :ivar cors: The set of CORS rules.
    :vartype cors: ~azure.table.models.CorsRule
    """

    _attribute_map = {
        "logging": {"key": "Logging", "type": "Logging"},
        "hour_metrics": {"key": "HourMetrics", "type": "Metrics"},
        "minute_metrics": {"key": "MinuteMetrics", "type": "Metrics"},
        "cors": {
            "key": "Cors",
            "type": "[CorsRule]",
            "xml": {"name": "Cors", "wrapped": True, "itemsName": "CorsRule"},
        },
    }
    _xml_map = {"name": "StorageServiceProperties"}

    def __init__(
        self,
        *,
        logging: Optional["_models.Logging"] = None,
        hour_metrics: Optional["_models.Metrics"] = None,
        minute_metrics: Optional["_models.Metrics"] = None,
        cors: Optional[List["_models.CorsRule"]] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword logging: Azure Analytics Logging settings.
        :paramtype logging: ~azure.table.models.Logging
        :keyword hour_metrics: A summary of request statistics grouped by API in hourly aggregates for
         tables.
        :paramtype hour_metrics: ~azure.table.models.Metrics
        :keyword minute_metrics: A summary of request statistics grouped by API in minute aggregates
         for tables.
        :paramtype minute_metrics: ~azure.table.models.Metrics
        :keyword cors: The set of CORS rules.
        :paramtype cors: ~azure.table.models.CorsRule
        """
        super().__init__(**kwargs)
        self.logging = logging
        self.hour_metrics = hour_metrics
        self.minute_metrics = minute_metrics
        self.cors = cors


class TableServiceStats(_serialization.Model):
    """Stats for the service.

    :ivar geo_replication: Geo-Replication information for the Secondary Storage Service.
    :vartype geo_replication: ~azure.table.models.GeoReplication
    """

    _attribute_map = {
        "geo_replication": {"key": "GeoReplication", "type": "GeoReplication"},
    }
    _xml_map = {"name": "StorageServiceStats"}

    def __init__(self, *, geo_replication: Optional["_models.GeoReplication"] = None, **kwargs: Any) -> None:
        """
        :keyword geo_replication: Geo-Replication information for the Secondary Storage Service.
        :paramtype geo_replication: ~azure.table.models.GeoReplication
        """
        super().__init__(**kwargs)
        self.geo_replication = geo_replication
